# Copyright 2019-2020 SURF.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from urllib.parse import urlparse

from fastapi import WebSocket, status
from fastapi.exceptions import HTTPException
from structlog import get_logger

from orchestrator.security import authenticate_websocket, authorize_websocket
from orchestrator.websocket.managers.broadcast_websocket_manager import BroadcastWebsocketManager
from orchestrator.websocket.managers.memory_websocket_manager import MemoryWebsocketManager

logger = get_logger(__name__)


class WebSocketManager:
    _backend: MemoryWebsocketManager | BroadcastWebsocketManager

    def __init__(self, websockets_enabled: bool, broadcast_url: str):
        self.enabled = websockets_enabled
        self.broadcaster_type = urlparse(broadcast_url).scheme
        self.connected = False
        if self.broadcaster_type in ("redis", "rediss"):
            self._backend = BroadcastWebsocketManager(broadcast_url)
        else:
            self._backend = MemoryWebsocketManager()

    async def authorize(self, websocket: WebSocket, token: str) -> dict | None:
        try:
            user = await authenticate_websocket(websocket=websocket, token=token)
            if user:
                await authorize_websocket(websocket, user)
        except HTTPException as e:
            return {"error": vars(e)}
        return None

    async def connect_redis(self) -> None:
        if not self.connected:
            await self._backend.connect_redis()
            self.connected = True

    async def disconnect_redis(self) -> None:
        if self.connected:
            await self._backend.disconnect_redis()
            self.connected = False

    async def connect(self, websocket: WebSocket, channel: str) -> None:
        await self._backend.connect(websocket, channel)

    async def disconnect(
        self, websocket: WebSocket, code: int = status.WS_1000_NORMAL_CLOSURE, reason: dict | str | None = None
    ) -> None:
        await self._backend.disconnect(websocket, code, reason)

    async def disconnect_all(self) -> None:
        await self._backend.disconnect_all()

    async def broadcast_data(self, channels: list[str], data: dict) -> None:
        await self._backend.broadcast_data(channels, data)



# Copyright 2019-2020 SURF.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from uuid import uuid4

import pytest
import sqlalchemy as sa
import structlog

from orchestrator.db import db, init_database
from orchestrator.migrations.helpers import (
    create_product_blocks,
    create_products,
    create_workflow,
    delete_workflow,
    ensure_default_workflows,
)
from orchestrator.settings import AppSettings
from orchestrator.targets import Target

logger = structlog.get_logger(__name__)


product_name = "Test Product"
product_block_name = "Test Product Block"

new_products = {
    product_name: {
        "product_id": uuid4(),
        "product_type": "TestProduct",
        "description": product_name,
        "tag": "TESTP",
        "status": "active",
        "product_block_ids": [],
    },
}

new_product_blocks = {
    product_block_name: {
        "product_block_id": uuid4(),
        "description": product_block_name,
        "tag": "TESTP",
        "status": "active",
        "resource_types": {
            "an_int": "An integer",
            "a_str": "A string",
            "a_bool": "A bool",
            "an_uuid": "An UUID",
            "an_ipv4": "An IPv4",
            "an_ipv6": "An IPv6",
        },
    },
}

new_workflows = [
    {
        "name": "create_test_product",
        "target": Target.CREATE,
        "description": "Create Test Product",
        "product_type": "TestProduct",
    },
]


@pytest.fixture(scope="session")
def new_test_product():
    settings = AppSettings()
    init_database(settings)
    conn = db.engine.connect()

    product_block_ids = create_product_blocks(conn, new_product_blocks)
    new_products[product_name]["product_block_ids"] = product_block_ids.values()
    create_products(conn, new_products)
    for workflow in new_workflows:
        create_workflow(conn, workflow)
    ensure_default_workflows(conn)

    yield None

    for workflow in new_workflows:
        delete_workflow(conn, workflow["name"])

    conn.execute(
        sa.text(
            "delete from processes_subscriptions where subscription_id in (select subscription_id from subscriptions where product_id in (select product_id from products where name = :name))"
        ),
        name=product_name,
    )
    conn.execute(
        sa.text("delete from subscriptions where product_id in (select product_id from products where name = :name)"),
        name=product_name,
    )

    conn.execute(sa.text("DELETE FROM products WHERE name = :name"), name=product_name)

    for product_block_name in new_product_blocks.keys():
        conn.execute(sa.text("DELETE FROM product_blocks WHERE name = :name"), name=product_block_name)

    db.session.flush()



import datetime
import os
import typing
from contextlib import closing
from typing import Any, cast

import pytest
import requests
import structlog
from alembic import command
from alembic.config import Config
from pydantic import BaseModel as PydanticBaseModel
from redis import Redis
from sqlalchemy import create_engine, select, text
from sqlalchemy.engine.url import make_url
from sqlalchemy.orm.scoping import scoped_session
from sqlalchemy.orm.session import sessionmaker
from starlette.testclient import TestClient
from urllib3_mock import Responses

from orchestrator import OrchestratorCore
from orchestrator.db import (
    ProductBlockTable,
    ProductTable,
    ResourceTypeTable,
    SubscriptionCustomerDescriptionTable,
    SubscriptionMetadataTable,
    WorkflowTable,
    db,
)
from orchestrator.db.database import ENGINE_ARGUMENTS, SESSION_ARGUMENTS, BaseModel, Database, SearchQuery
from orchestrator.domain import SUBSCRIPTION_MODEL_REGISTRY, SubscriptionModel
from orchestrator.domain.base import ProductBlockModel
from orchestrator.services.translations import generate_translations
from orchestrator.settings import app_settings
from orchestrator.types import SubscriptionLifecycle
from orchestrator.utils.json import json_dumps
from pydantic_forms.core import FormPage
from test.unit_tests.fixtures.processes import mocked_processes, mocked_processes_resumeall, test_workflow  # noqa: F401
from test.unit_tests.fixtures.products.product_blocks.product_block_list_nested import (  # noqa: F401
    test_product_block_list_nested,
    test_product_block_list_nested_db_in_use_by_block,
)
from test.unit_tests.fixtures.products.product_blocks.product_block_one import (  # noqa: F401
    test_product_block_one,
    test_product_block_one_db,
)
from test.unit_tests.fixtures.products.product_blocks.product_block_one_nested import (  # noqa: F401
    test_product_block_one_nested,
    test_product_block_one_nested_db_in_use_by_block,
)
from test.unit_tests.fixtures.products.product_blocks.product_block_with_list_union import (  # noqa: F401
    test_product_block_with_list_union,
    test_product_block_with_list_union_db,
)
from test.unit_tests.fixtures.products.product_blocks.product_block_with_union import (  # noqa: F401
    test_product_block_with_union,
    test_product_block_with_union_db,
)
from test.unit_tests.fixtures.products.product_blocks.product_sub_block_one import (  # noqa: F401
    test_product_sub_block_one,
    test_product_sub_block_one_db,
)
from test.unit_tests.fixtures.products.product_blocks.product_sub_block_two import (  # noqa: F401
    test_product_sub_block_two,
    test_product_sub_block_two_db,
)
from test.unit_tests.fixtures.products.product_types.product_type_list_nested import (  # noqa: F401
    test_product_list_nested,
    test_product_model_list_nested,
    test_product_type_list_nested,
)
from test.unit_tests.fixtures.products.product_types.product_type_list_union import (  # noqa: F401
    test_product_list_union,
    test_product_type_list_union,
)
from test.unit_tests.fixtures.products.product_types.product_type_list_union_overlap import (  # noqa: F401
    sub_list_union_overlap_subscription_1,
    test_product_list_union_overlap,
    test_product_type_list_union_overlap,
)
from test.unit_tests.fixtures.products.product_types.product_type_one import (  # noqa: F401
    product_one_subscription_1,
    test_product_model,
    test_product_one,
    test_product_type_one,
)
from test.unit_tests.fixtures.products.product_types.product_type_one_nested import (  # noqa: F401
    test_product_model_nested,
    test_product_one_nested,
    test_product_type_one_nested,
)
from test.unit_tests.fixtures.products.product_types.product_type_sub_list_union import (  # noqa: F401
    product_sub_list_union_subscription_1,
    test_product_sub_list_union,
    test_product_type_sub_list_union,
)
from test.unit_tests.fixtures.products.product_types.product_type_sub_one import (  # noqa: F401
    sub_one_subscription_1,
    test_product_sub_one,
    test_product_type_sub_one,
)
from test.unit_tests.fixtures.products.product_types.product_type_sub_two import (  # noqa: F401
    sub_two_subscription_1,
    test_product_sub_two,
    test_product_type_sub_two,
)
from test.unit_tests.fixtures.products.product_types.product_type_sub_union import (  # noqa: F401
    test_union_sub_product,
    test_union_type_sub_product,
)
from test.unit_tests.fixtures.products.product_types.product_type_union import (  # noqa: F401
    test_union_product,
    test_union_type_product,
)
from test.unit_tests.fixtures.products.resource_types import (  # noqa: F401
    resource_type_enum,
    resource_type_int,
    resource_type_int_2,
    resource_type_list,
    resource_type_str,
)
from test.unit_tests.workflows import WorkflowInstanceForTests
from test.unit_tests.workflows.shared.test_validate_subscriptions import validation_workflow

logger = structlog.getLogger(__name__)

CUSTOMER_ID: str = "2f47f65a-0911-e511-80d0-005056956c1a"


def run_migrations(db_uri: str) -> None:
    """Configure the alembic context and run the migrations.

    Each test will start with a clean database. This a heavy operation but ensures that our database is clean and
    tests run within their own context.

    Args:
        db_uri: The database uri configuration to run the migration on.

    Returns:
        None

    """
    path = os.path.join(os.path.dirname(os.path.realpath(__file__)))
    os.environ["DATABASE_URI"] = db_uri
    app_settings.DATABASE_URI = db_uri  # type: ignore
    alembic_cfg = Config(file_=os.path.join(path, "../../orchestrator/migrations/alembic.ini"))
    alembic_cfg.set_main_option("script_location", os.path.join(path, "../../orchestrator/migrations"))
    alembic_cfg.set_main_option(
        "version_locations",
        f"{os.path.join(path, '../../orchestrator/migrations/versions/schema')}",
    )
    alembic_cfg.set_main_option("sqlalchemy.url", db_uri)
    command.upgrade(alembic_cfg, "heads")


@pytest.fixture(scope="session")
def db_uri(worker_id):
    """Ensure each pytest thread has its database.

    When running tests with the -j option make sure each test worker is isolated within its own database.

    Args:
        worker_id: the worker id

    Returns:
        Database uri to be used in the test thread

    """
    database_uri = os.environ.get("DATABASE_URI", "postgresql://nwa:nwa@localhost/orchestrator-core-test")
    if worker_id == "master":
        # pytest is being run without any workers
        return database_uri
    url = make_url(database_uri)
    if hasattr(url, "set"):
        url = url.set(database=f"{url.database}-{worker_id}")
    else:
        url.database = f"{url.database}-{worker_id}"
    return url.render_as_string(hide_password=False)


@pytest.fixture(scope="session")
def database(db_uri):
    """Create database and run migrations and cleanup afterward.

    Args:
        db_uri: fixture for providing the application context and an initialized database. Although specifying this
            as an explicit parameter is redundant due to `fastapi_app`'s autouse setting, we have made the dependency
            explicit here for the purpose of documentation.

    """
    db.update(Database(db_uri))
    url = make_url(db_uri)
    db_to_create = url.database
    if hasattr(url, "set"):
        url = url.set(database="postgres")
    else:
        url.database = "postgres"
    engine = create_engine(url)
    with closing(engine.connect()) as conn:
        conn.execute(text("COMMIT;"))
        conn.execute(text(f'DROP DATABASE IF EXISTS "{db_to_create}";'))
        conn.execute(text("COMMIT;"))
        conn.execute(text(f'CREATE DATABASE "{db_to_create}";'))

    run_migrations(db_uri)
    db.wrapped_database.engine = create_engine(db_uri, **ENGINE_ARGUMENTS)

    try:
        yield
    finally:
        db.wrapped_database.engine.dispose()
        with closing(engine.connect()) as conn:
            conn.execute(text("COMMIT;"))
            conn.execute(text(f'DROP DATABASE IF EXISTS "{db_to_create}";'))


@pytest.fixture(autouse=True)
def db_session(database):
    """Ensure tests are run in a transaction with automatic rollback.

    This implementation creates a connection and transaction before yielding to the test function. Any transactions
    started and committed from within the test will be tied to this outer transaction. From the test function's
    perspective it looks like everything will indeed be committed; allowing for queries on the database to be
    performed to see if functions under test have persisted their changes to the database correctly. However once
    the test function returns this fixture will clean everything up by rolling back the outer transaction; leaving the
    database in a known state (=empty except what migrations have added as the initial state).

    Args:
        database: fixture for providing an initialized database.

    """
    with closing(db.wrapped_database.engine.connect()) as test_connection:
        db.wrapped_database.session_factory = sessionmaker(**SESSION_ARGUMENTS, bind=test_connection)
        db.wrapped_database.scoped_session = scoped_session(db.session_factory, db._scopefunc)
        BaseModel.set_query(cast(SearchQuery, db.wrapped_database.scoped_session.query_property()))

        trans = test_connection.begin()
        try:
            yield
        finally:
            if not trans._deactivated_from_connection:
                trans.rollback()


@pytest.fixture(scope="session", autouse=True)
def fastapi_app(database, db_uri):
    from oauth2_lib.settings import oauth2lib_settings

    oauth2lib_settings.OAUTH2_ACTIVE = False
    oauth2lib_settings.ENVIRONMENT_IGNORE_MUTATION_DISABLED = ["local", "TESTING"]
    app_settings.DATABASE_URI = db_uri
    app = OrchestratorCore(base_settings=app_settings)
    # Start ProcessDataBroadcastThread to test websocket_manager with memory backend
    app.broadcast_thread.start()
    yield app
    app.broadcast_thread.stop()
    oauth2lib_settings.OAUTH2_ACTIVE = True


@pytest.fixture(scope="session")
def test_client(fastapi_app):
    class JsonTestClient(TestClient):
        def request(  # type: ignore
            self,
            method: str,
            url: str,
            data: Any | None = None,
            headers: typing.MutableMapping[str, str] | None = None,
            json: typing.Any = None,
            **kwargs: Any,
        ) -> requests.Response:
            if json is not None:
                if headers is None:
                    headers = {}
                data = json_dumps(json).encode()
                headers["Content-Type"] = "application/json"

            return super().request(method, url, data=data, headers=headers, **kwargs)  # type: ignore

    return JsonTestClient(fastapi_app)


@pytest.fixture(autouse=True)
def responses():
    responses_mock = Responses("requests.packages.urllib3")

    def _find_request(call):
        mock_url = responses_mock._find_match(call.request)
        if not mock_url:
            raise Exception(f"Call not mocked: {call.request}")
        return mock_url

    def _to_tuple(url_mock):
        return (url_mock["url"], url_mock["method"], url_mock["match_querystring"])

    with responses_mock:
        yield responses_mock

        mocked_urls = map(_to_tuple, responses_mock._urls)
        used_urls = map(_to_tuple, map(_find_request, responses_mock.calls))
        not_used = set(mocked_urls) - set(used_urls)
        if not_used:
            pytest.fail(f"Found unused responses mocks: {not_used}", pytrace=False)


@pytest.fixture(scope="session", autouse=True)
def test_form_translations(worker_id):
    """Some voodoo to check for each form during test if the translations are complete."""

    translations = generate_translations("en-GB")["forms"]["fields"]
    used_translations = set()

    # In order to properly wrap a classmethod we need to do special stuff
    old_init_subclass = FormPage.__dict__["__pydantic_init_subclass__"]

    # Wrap a form function that is certain to be called to extract the used form fields
    @classmethod
    def init_subclass_wrapper(cls: type[PydanticBaseModel], *args, **kwargs: Any) -> None:
        # Skip forms in test modules
        if "test" not in cls.__module__:
            for field_name in cls.model_fields:
                used_translations.add(field_name)
                if field_name not in translations and f"{field_name}_accept" not in translations:
                    pytest.fail(f"Missing translation for field {field_name} in  {cls.__name__}")

        # Because the original is a classmethod we need to conform to the descriptor protocol
        return old_init_subclass.__get__(None, cls)(*args, **kwargs)

    FormPage.__pydantic_init_subclass__ = init_subclass_wrapper
    try:
        yield
    finally:
        # unwrapp and check if all translations are actually used
        FormPage.__pydantic_init_subclass__ = old_init_subclass

        # This check only works when you run without python-xdist because we need one single session
        # TODO this does not work reliable yet
        # if worker_id == "master":
        #     unused_keys = set()
        #     for trans_key in translations:
        #         if (
        #             not trans_key.endswith("_info")
        #             and not trans_key.endswith("_accept")
        #             and not trans_key.endswith("_fields")
        #             and trans_key not in used_translations
        #             and f"{trans_key}_accept" not in used_translations
        #         ):
        #             unused_keys.add(trans_key)

        #     if unused_keys:
        #         pytest.fail(f"found unused translations: {sorted(unused_keys)}", pytrace=False)


@pytest.fixture
def generic_resource_type_1():
    rt = ResourceTypeTable(description="Resource Type one", resource_type="rt_1")
    db.session.add(rt)
    db.session.commit()

    return rt


@pytest.fixture
def generic_resource_type_2():
    rt = ResourceTypeTable(description="Resource Type two", resource_type="rt_2")
    db.session.add(rt)
    db.session.commit()
    return rt


@pytest.fixture
def generic_resource_type_3():
    rt = ResourceTypeTable(description="Resource Type three", resource_type="rt_3")
    db.session.add(rt)
    db.session.commit()

    return rt


@pytest.fixture
def generic_product_block_1(generic_resource_type_1):
    pb = ProductBlockTable(
        name="PB_1",
        description="Generic Product Block 1",
        tag="PB1",
        status="active",
        resource_types=[generic_resource_type_1],
        created_at=datetime.datetime.fromisoformat("2023-05-24T00:00:00+00:00"),
    )
    db.session.add(pb)
    db.session.commit()
    return pb


@pytest.fixture
def generic_product_block_2(generic_resource_type_2, generic_resource_type_3):
    pb = ProductBlockTable(
        name="PB_2",
        description="Generic Product Block 2",
        tag="PB2",
        status="active",
        resource_types=[generic_resource_type_2, generic_resource_type_3],
        created_at=datetime.datetime.fromisoformat("2023-05-24T00:00:00+00:00"),
    )
    db.session.add(pb)
    db.session.commit()
    return pb


@pytest.fixture
def generic_product_block_3(generic_resource_type_2):
    pb = ProductBlockTable(
        name="PB_3",
        description="Generic Product Block 3",
        tag="PB3",
        status="active",
        resource_types=[generic_resource_type_2],
        created_at=datetime.datetime.fromisoformat("2023-05-24T00:00:00+00:00"),
    )
    db.session.add(pb)
    db.session.commit()
    return pb


@pytest.fixture
def generic_product_1(generic_product_block_1, generic_product_block_2):
    workflow = db.session.scalar(select(WorkflowTable).where(WorkflowTable.name == "modify_note"))
    p = ProductTable(
        name="Product 1",
        description="Generic Product One",
        product_type="Generic",
        status="active",
        tag="GEN1",
        product_blocks=[generic_product_block_1, generic_product_block_2],
        workflows=[workflow],
    )
    db.session.add(p)
    db.session.commit()
    return p


@pytest.fixture
def generic_product_2(generic_product_block_3):
    workflow = db.session.scalar(select(WorkflowTable).where(WorkflowTable.name == "modify_note"))

    p = ProductTable(
        name="Product 2",
        description="Generic Product Two",
        product_type="Generic",
        status="active",
        tag="GEN2",
        product_blocks=[generic_product_block_3],
        workflows=[workflow],
    )
    db.session.add(p)
    db.session.commit()
    return p


@pytest.fixture
def generic_product_3(generic_product_block_2):
    p = ProductTable(
        name="Product 3",
        description="Generic Product Three",
        product_type="Generic",
        status="active",
        tag="GEN3",
        product_blocks=[generic_product_block_2],
    )
    db.session.add(p)
    db.session.commit()
    return p


@pytest.fixture
def generic_product_block_type_1(generic_product_block_1):
    class GenericProductBlockOneInactive(ProductBlockModel, product_block_name="PB_1"):
        rt_1: str | None = None

    class GenericProductBlockOne(GenericProductBlockOneInactive, lifecycle=[SubscriptionLifecycle.ACTIVE]):
        rt_1: str

    return GenericProductBlockOneInactive, GenericProductBlockOne


@pytest.fixture
def generic_product_block_type_2(generic_product_block_2):
    class GenericProductBlockTwoInactive(ProductBlockModel, product_block_name="PB_2"):
        rt_2: int | None = None
        rt_3: str | None = None

    class GenericProductBlockTwo(GenericProductBlockTwoInactive, lifecycle=[SubscriptionLifecycle.ACTIVE]):
        rt_2: int
        rt_3: str

    return GenericProductBlockTwoInactive, GenericProductBlockTwo


@pytest.fixture
def generic_product_block_type_3(generic_product_block_3):
    class GenericProductBlockThreeInactive(ProductBlockModel, product_block_name="PB_3"):
        rt_2: int | None = None

    class GenericProductBlockThree(GenericProductBlockThreeInactive, lifecycle=[SubscriptionLifecycle.ACTIVE]):
        rt_2: int

    return GenericProductBlockThreeInactive, GenericProductBlockThree


@pytest.fixture
def generic_product_type_1(generic_product_1, generic_product_block_type_1, generic_product_block_type_2):
    GenericProductBlockOneInactive, GenericProductBlockOne = generic_product_block_type_1
    GenericProductBlockTwoInactive, GenericProductBlockTwo = generic_product_block_type_2
    # Test Product domain models

    class GenericProductOneInactive(SubscriptionModel, is_base=True):
        pb_1: GenericProductBlockOneInactive
        pb_2: GenericProductBlockTwoInactive

    class GenericProductOne(GenericProductOneInactive, lifecycle=[SubscriptionLifecycle.ACTIVE]):
        pb_1: GenericProductBlockOne
        pb_2: GenericProductBlockTwo

    SUBSCRIPTION_MODEL_REGISTRY["Product 1"] = GenericProductOne

    yield GenericProductOneInactive, GenericProductOne

    del SUBSCRIPTION_MODEL_REGISTRY["Product 1"]


@pytest.fixture
def generic_product_type_2(generic_product_2, generic_product_block_type_3):
    GenericProductBlockThreeInactive, GenericProductBlockThree = generic_product_block_type_3

    class GenericProductTwoInactive(SubscriptionModel, is_base=True):
        pb_3: GenericProductBlockThreeInactive

    class GenericProductTwo(GenericProductTwoInactive, lifecycle=[SubscriptionLifecycle.ACTIVE]):
        pb_3: GenericProductBlockThree

    SUBSCRIPTION_MODEL_REGISTRY["Product 2"] = GenericProductTwo

    yield GenericProductTwoInactive, GenericProductTwo

    del SUBSCRIPTION_MODEL_REGISTRY["Product 2"]


@pytest.fixture
def product_type_1_subscription_factory(generic_product_1, generic_product_type_1):
    def subscription_create(
        description="Generic Subscription One",
        start_date="2023-05-24T00:00:00+00:00",
        rt_1="Value1",
        rt_2=42,
        rt_3="Value2",
    ):
        GenericProductOneInactive, _ = generic_product_type_1
        gen_subscription = GenericProductOneInactive.from_product_id(
            generic_product_1.product_id, customer_id=CUSTOMER_ID, insync=True
        )
        gen_subscription.pb_1.rt_1 = rt_1
        gen_subscription.pb_2.rt_2 = rt_2
        gen_subscription.pb_2.rt_3 = rt_3
        gen_subscription = SubscriptionModel.from_other_lifecycle(gen_subscription, SubscriptionLifecycle.ACTIVE)
        gen_subscription.description = description
        gen_subscription.start_date = start_date
        gen_subscription.save()

        gen_subscription_metadata = SubscriptionMetadataTable()
        gen_subscription_metadata.subscription_id = gen_subscription.subscription_id
        gen_subscription_metadata.metadata_ = {"description": "Some metadata description"}
        db.session.add(gen_subscription_metadata)
        db.session.commit()
        return str(gen_subscription.subscription_id)

    return subscription_create


@pytest.fixture
def product_type_1_subscriptions_factory(product_type_1_subscription_factory):
    def subscriptions_create(amount=1):
        return [
            product_type_1_subscription_factory(
                description=f"Subscription {i}",
                start_date=(
                    datetime.datetime.fromisoformat("2023-05-24T00:00:00+00:00") + datetime.timedelta(days=i)
                ).replace(tzinfo=datetime.UTC),
            )
            for i in range(0, amount)
        ]

    return subscriptions_create


@pytest.fixture
def generic_subscription_1(product_type_1_subscription_factory):
    return product_type_1_subscription_factory()


@pytest.fixture
def generic_subscription_2(generic_product_2, generic_product_type_2):
    GenericProductTwoInactive, _ = generic_product_type_2
    gen_subscription = GenericProductTwoInactive.from_product_id(
        generic_product_2.product_id, customer_id=CUSTOMER_ID, insync=True
    )
    gen_subscription.pb_3.rt_2 = 42
    gen_subscription = SubscriptionModel.from_other_lifecycle(gen_subscription, SubscriptionLifecycle.ACTIVE)
    gen_subscription.description = "Generic Subscription One"
    gen_subscription.save()
    db.session.commit()

    return str(gen_subscription.subscription_id)


@pytest.fixture
def validation_workflow_instance():
    with WorkflowInstanceForTests(validation_workflow, "validation_workflow") as ctx:
        yield ctx


@pytest.fixture
def make_customer_description():
    def customer_description(subscription_id, customer_id, description):
        model = SubscriptionCustomerDescriptionTable(
            subscription_id=subscription_id, customer_id=customer_id, description=description
        )
        db.session.add(model)
        db.session.commit()
        return model

    return customer_description


@pytest.fixture
def cache_fixture(monkeypatch):
    """Fixture to enable domain model caching and cleanup keys added to the list."""
    with monkeypatch.context() as m:
        m.setattr(app_settings, "CACHE_DOMAIN_MODELS", True)
        cache = Redis.from_url(str(app_settings.CACHE_URI))
        # Clear cache before using this fixture
        cache.flushdb()

        to_cleanup = []

        yield to_cleanup

        for key in to_cleanup:
            try:
                cache.delete(key)
            except Exception as exc:
                print("failed to delete cache key", key, str(exc))  # noqa: T001, T201


def do_refresh_subscriptions_search_view():
    db.session.execute(text("REFRESH MATERIALIZED VIEW subscriptions_search"))


@pytest.fixture
def refresh_subscriptions_search_view():
    do_refresh_subscriptions_search_view()



import uuid

import pytest
from sqlalchemy import Column, Integer, String, func, select

from orchestrator.db import db
from orchestrator.db.database import BaseModel
from orchestrator.db.helpers import to_sql_string
from orchestrator.utils.search_query import Lexer, ParseError, Parser, TSQueryVisitor, create_sqlalchemy_select


def _parse_tree_and_tsquery(q: str) -> tuple[tuple, str]:
    tokens = Lexer(q).lex()
    tree = Parser(tokens).parse()
    tsquery = TSQueryVisitor.visit(tree)
    return tree, tsquery


def test_parse_simple_query():
    q = "simple query with words"
    parse_tree, tsquery = _parse_tree_and_tsquery(q)
    assert parse_tree == (
        "Query",
        [("AndExpression", [("Word", "simple"), ("Word", "query"), ("Word", "with"), ("Word", "words")])],
    )
    assert tsquery == "simple & query & with & words"


def test_parse_query_with_phrase():
    q = 'word1 "some consecutive words" end'
    parse_tree, tsquery = _parse_tree_and_tsquery(q)
    assert parse_tree == (
        "Query",
        [
            (
                "AndExpression",
                [
                    ("Word", "word1"),
                    ("Phrase", [("Word", "some"), ("Word", "consecutive"), ("Word", "words")]),
                    ("Word", "end"),
                ],
            )
        ],
    )
    assert tsquery == "word1 & some <-> consecutive <-> words & end"


@pytest.mark.parametrize(
    "query, msg",
    [
        ("a )", "Right paren before left paren"),
        ('tag:""', "Empty Value in KVTerm"),
        ("tag:()", "Empty ValueGroup in KVTerm"),
        ("tag:(a | *b(", "Unexpected token in ValueGroup"),
        ('this "is unbalanced', "Missing closing quote"),
        ("a (unclosed group", "Missing closing parenthesis"),
    ],
)
def test_parse_errors(query, msg):
    with pytest.raises(ParseError):
        _parse_tree_and_tsquery(query)


def test_parse_edge_cases():
    # Just test whether this parses without error
    q = "aa*bb@cc?dd>ee'ff_g<g[hh]ij$kk%;\\=+12`3€crazy!text kk|ll:m"
    parse_tree, tsquery = _parse_tree_and_tsquery(q)
    assert tsquery.startswith("aa:*")
    assert tsquery.endswith("| ll <-> m")


@pytest.mark.parametrize(
    "query,expected_tsquery",
    [
        ("basic query example", "basic & query & example"),
        ("example status:(suspended | terminated)", "example & status <-> (suspended | terminated)"),
        (
            "subscription_id:849fb72a-4366-4d41-9d37-3821d73e4dd4",
            "(subscription <-> id) <-> (849fb72a <-> 4366 <-> 4d41 <-> 9d37 <-> 3821d73e4dd4)",
        ),
        ("something aaa_bbb else", "something & (aaa <-> bbb) & else"),
    ],
)
def test_common_queries(query, expected_tsquery):
    _, tsquery = _parse_tree_and_tsquery(query)
    assert tsquery == expected_tsquery


def test_parse_complex_query():
    q = "".join(
        [
            '"phrase1 phrase2 phrase3" word1 word2 prefixword* field1:(val1 | val2* | "val3 val4*")',
            " | ",
            '"phrase21 phrase22" ((klm tag:lp) | (sinica tag:fw)) -("not this" "or this") something else',
        ]
    )

    parse_tree, tsquery = _parse_tree_and_tsquery(q)
    assert len(parse_tree[1]) == 2
    first_subexpression = parse_tree[1][0]
    second_subexpression = parse_tree[1][1]
    assert len(first_subexpression[1]) == 5, "subexpression 1 has 5 terms"
    assert len(second_subexpression[1]) == 5, "subexpression 2 has 5 terms"
    assert tsquery == "".join(
        [
            "phrase1 <-> phrase2 <-> phrase3 & word1 & word2 & prefixword:* & field1 <-> (val1 | val2:* | val3 <-> val4:*)",
            " | ",
            "phrase21 <-> phrase22 & ((klm & tag <-> lp) | (sinica & tag <-> fw))",
            " & !(not <-> this & or <-> this) & something & else",
        ]
    )


def test_query_with_underscores():
    q = 'word_with_underscores prefix_word* key_value:term_1 key_value2:(val1 | "val2" | val3*)'
    parse_tree, tsquery = _parse_tree_and_tsquery(q)
    assert (
        tsquery
        == "(word <-> with <-> underscores) & (prefix <-> word:*) & (key <-> value) <-> (term <-> 1) & (key <-> value2) <-> (val1 | val2 | val3:*)"
    )


def test_query_kv_camelcasing():
    q = "productTag:myTag"
    q2 = "product_tag:myTag"
    assert _parse_tree_and_tsquery(q)[1] == _parse_tree_and_tsquery(q2)[1]


@pytest.mark.parametrize(
    "search_query,substrings",
    [
        ("name:Bob", ["WHERE my_table.name LIKE 'bob'"]),
        ("empty:column", ["WHERE false"]),
        ("bob| joe", ["WHERE my_table.name LIKE 'bob'", "UNION SELECT", "WHERE my_table.name LIKE 'joe'"]),
        ("(bob) | -(joe)", ["WHERE my_table.name LIKE 'bob'", "UNION SELECT", "WHERE my_table.name LIKE 'joe'"]),
        ("-(joe)", ["WITH", "LEFT OUTER JOIN", ".id IS NULL"]),
    ],
)
def test_sqlalchemy_visitor(search_query, substrings):
    class MyTable(BaseModel):
        __tablename__ = "my_table"
        id = Column(Integer, primary_key=True)
        name = Column(String)

    mappings = {"name": lambda node: MyTable.name.like(node[1].lower())}
    stmt = select(MyTable)
    stmt = create_sqlalchemy_select(stmt, search_query, mappings, MyTable, MyTable.id)
    for substring in substrings:
        assert substring in to_sql_string(stmt)

    MyTable.metadata.clear()


@pytest.mark.parametrize("i", range(100))
def test_uuid_queries(i):
    # Tests whether a text search vector containing a UUID can be successfully queried
    # This used to fail sometimes in Postgres <14 versions.

    # TODO: Remove when support for Postgres <14 is dropped
    uid = uuid.uuid4()

    query = f"subscription_id:{uid}"
    tsv = func.to_tsvector("simple", func.replace(query, "-", "_"))

    _, ts_query = _parse_tree_and_tsquery(query)

    stmt = select(tsv.op("@@")(func.to_tsquery("simple", ts_query)))

    matches = db.session.scalar(stmt)
    assert matches



"""Add example2 product.

Revision ID: 85be1c80731c
Revises: 59e1199aff7f
Create Date: 2024-02-20 21:01:46.178206

"""

from uuid import uuid4

from alembic import op
from orchestrator.migrations.helpers import create, create_workflow, delete, delete_workflow, ensure_default_workflows
from orchestrator.targets import Target

# revision identifiers, used by Alembic.
revision = "85be1c80731c"
down_revision = "59e1199aff7f"
branch_labels = None
depends_on = None

new_products = {
    "products": {
        "example2": {
            "product_id": uuid4(),
            "product_type": "Example2",
            "description": "Product example 2",
            "tag": "EXAMPLE2",
            "status": "active",
            "root_product_block": "Example2",
            "fixed_inputs": {},
        },
    },
    "product_blocks": {
        "Example2": {
            "product_block_id": uuid4(),
            "description": "Example 2 root product block",
            "tag": "EXAMPLE2",
            "status": "active",
            "resources": {
                "example_int_enum_2": "Example 2 int enum",
            },
            "depends_on_block_relations": [],
        },
    },
    "workflows": {},
}

new_workflows = [
    {
        "name": "create_example2",
        "target": Target.CREATE,
        "description": "Create example2",
        "product_type": "Example2",
    },
    {
        "name": "modify_example2",
        "target": Target.MODIFY,
        "description": "Modify example2",
        "product_type": "Example2",
    },
    {
        "name": "terminate_example2",
        "target": Target.TERMINATE,
        "description": "Terminate example2",
        "product_type": "Example2",
    },
    {
        "name": "validate_example2",
        "target": Target.SYSTEM,
        "description": "Validate example2",
        "product_type": "Example2",
    },
]


def upgrade() -> None:
    conn = op.get_bind()
    create(conn, new_products)
    for workflow in new_workflows:
        create_workflow(conn, workflow)
    ensure_default_workflows(conn)


def downgrade() -> None:
    conn = op.get_bind()
    for workflow in new_workflows:
        delete_workflow(conn, workflow["name"])

    delete(conn, new_products)



import pytest
from orchestrator.types import SubscriptionLifecycle

from products.product_types.example4 import Example4
from test.unit_tests.workflows import assert_complete, extract_state, run_workflow


@pytest.mark.workflow
def test_happy_flow(responses, example4_subscription):
    # when

    # TODO: insert mocks here if needed

    result, _, _ = run_workflow("terminate_example4", [{"subscription_id": example4_subscription}, {}])

    # then

    assert_complete(result)
    state = extract_state(result)
    assert "subscription" in state

    # Check subscription in DB

    example4 = Example4.from_subscription(example4_subscription)
    assert example4.end_date is not None
    assert example4.status == SubscriptionLifecycle.TERMINATED



from datetime import timedelta
from uuid import uuid4

import pytest
from sqlalchemy import select

from orchestrator.db import ProcessStepTable, ProcessTable, WorkflowTable, db
from orchestrator.targets import Target
from orchestrator.utils.datetime import nowtz
from orchestrator.workflow import ProcessStatus
from test.unit_tests.workflows import assert_complete, assert_state, extract_state, run_workflow


@pytest.fixture
def task():
    three_weeks_ago = nowtz() - timedelta(weeks=3)
    two_weeks_ago = nowtz() - timedelta(weeks=2)
    state = {"foo": "bar"}

    generic_step = ProcessStepTable(name="generic-step", status="success", state=state)

    wf_old = WorkflowTable(workflow_id=uuid4(), name="nice and old task", target=Target.SYSTEM)
    task_old = ProcessTable(
        workflow_id=wf_old.workflow_id,
        last_status=ProcessStatus.COMPLETED,
        last_step="Awesome last step",
        started_at=three_weeks_ago,
        last_modified_at=two_weeks_ago,
        steps=[generic_step],
        is_task=True,
    )
    wf_new = WorkflowTable(workflow_id=uuid4(), name="nice and new task", target=Target.SYSTEM)

    task_new = ProcessTable(
        workflow_id=wf_new.workflow_id,
        last_status=ProcessStatus.COMPLETED,
        last_step="Awesome last step",
        started_at=three_weeks_ago,
        last_modified_at=nowtz(),
        steps=[generic_step],
        is_task=True,
    )
    wf = WorkflowTable(workflow_id=uuid4(), name="nice process", target=Target.SYSTEM)

    process = ProcessTable(
        workflow_id=wf.workflow_id,
        last_status=ProcessStatus.COMPLETED,
        last_step="Awesome last step",
        started_at=three_weeks_ago,
        last_modified_at=two_weeks_ago,
        steps=[generic_step],
        is_task=False,
    )
    db.session.add_all([wf_old, wf_new, wf, generic_step, task_old, task_new, process])
    db.session.commit()


@pytest.mark.workflow
def test_remove_tasks(task):
    result, process, step_log = run_workflow("task_clean_up_tasks", {})
    assert_complete(result)
    res = extract_state(result)
    state = {"process_id": res["process_id"], "reporter": "john.doe", "tasks_removed": 1}
    assert_state(result, state)

    processes = db.session.scalars(select(ProcessTable)).all()

    assert len(processes) == 3
    assert sorted(p.workflow.name for p in processes) == sorted(
        ["nice and new task", "nice process", "task_clean_up_tasks"]
    )



# Copyright 2019-2020 SURF.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from orchestrator.schedules.resume_workflows import run_resume_workflows
from orchestrator.schedules.scheduling import SchedulingFunction
from orchestrator.schedules.task_vacuum import vacuum_tasks
from orchestrator.schedules.validate_products import validate_products
from orchestrator.schedules.validate_subscriptions import validate_subscriptions

ALL_SCHEDULERS: list[SchedulingFunction] = [
    run_resume_workflows,
    vacuum_tasks,
    validate_subscriptions,
    validate_products,
]



# Copyright 2019-2020 SURF.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from datetime import datetime

from pydantic import BaseModel, ConfigDict


class OrchestratorBaseModel(BaseModel):
    # Unable to refactor in a neat way without using an ugly recursive function. Leaving Pydantic v1 functionality intact
    model_config = ConfigDict(
        json_encoders={
            datetime: lambda dt: dt.timestamp(),
        }
    )



import structlog
from orchestrator.domain import SubscriptionModel
from orchestrator.forms import FormPage
from orchestrator.forms.validators import CustomerId, Divider
from orchestrator.types import FormGenerator, State, SubscriptionLifecycle, UUIDstr
from orchestrator.workflow import StepList, begin, step
from orchestrator.workflows.steps import set_status
from orchestrator.workflows.utils import modify_workflow

from products.product_blocks.example2 import ExampleIntEnum2
from products.product_types.example2 import Example2, Example2Provisioning


def subscription_description(subscription: SubscriptionModel) -> str:
    """The suggested pattern is to implement a subscription service that generates a subscription specific
    description, in case that is not present the description will just be set to the product name.
    """
    return f"{subscription.product.name} subscription"


logger = structlog.get_logger(__name__)


def initial_input_form_generator(subscription_id: UUIDstr) -> FormGenerator:
    subscription = Example2.from_subscription(subscription_id)
    example2 = subscription.example2

    # TODO fill in additional fields if needed

    class ModifyExample2Form(FormPage):
        customer_id: CustomerId = subscription.customer_id  # type: ignore

        divider_1: Divider

        example_int_enum_2: ExampleIntEnum2 | None = example2.example_int_enum_2

    user_input = yield ModifyExample2Form
    user_input_dict = user_input.dict()

    return user_input_dict | {"subscription": subscription}


@step("Update subscription")
def update_subscription(
    subscription: Example2Provisioning,
    example_int_enum_2: ExampleIntEnum2 | None,
) -> State:
    # TODO: get all modified fields
    subscription.example2.example_int_enum_2 = example_int_enum_2

    return {"subscription": subscription}


@step("Update subscription description")
def update_subscription_description(subscription: Example2) -> State:
    subscription.description = subscription_description(subscription)
    return {"subscription": subscription}


additional_steps = begin


@modify_workflow("Modify example2", initial_input_form=initial_input_form_generator, additional_steps=additional_steps)
def modify_example2() -> StepList:
    return (
        begin
        >> set_status(SubscriptionLifecycle.PROVISIONING)
        >> update_subscription
        >> update_subscription_description
        # TODO add additional steps if needed
        >> set_status(SubscriptionLifecycle.ACTIVE)
    )



from unittest.mock import MagicMock

from strawberry.types.nodes import SelectedField

from orchestrator.graphql.utils import is_querying_page_data


def test_is_querying_page_data_with_pageinfo_fields():
    info = MagicMock(
        selected_fields=[
            SelectedField(
                name="subscriptions",
                directives={},
                arguments={},
                alias=None,
                selections=[
                    SelectedField(
                        name="pageInfo",
                        directives={},
                        arguments={},
                        alias=None,
                        selections=[
                            SelectedField(name="totalItems", directives={}, arguments={}, alias=None, selections=[]),
                        ],
                    ),
                ],
            ),
        ]
    )

    assert not is_querying_page_data(info)


def test_is_querying_page_data_with_page_fields():
    info = MagicMock(
        selected_fields=[
            SelectedField(
                name="subscriptions",
                directives={},
                arguments={},
                alias=None,
                selections=[
                    SelectedField(
                        name="page",
                        directives={},
                        arguments={},
                        alias=None,
                        selections=[
                            SelectedField(
                                name="subscription_id", directives={}, arguments={}, selections=[], alias=None
                            ),
                            SelectedField(name="description", directives={}, arguments={}, selections=[], alias=None),
                        ],
                    ),
                ],
            ),
        ]
    )

    assert is_querying_page_data(info)


def test_is_querying_page_data_with_page_and_paginfo_fields():
    info = MagicMock(
        selected_fields=[
            SelectedField(
                name="subscriptions",
                directives={},
                arguments={},
                selections=[
                    SelectedField(
                        name="pageInfo",
                        directives={},
                        arguments={},
                        alias=None,
                        selections=[
                            SelectedField(name="totalItems", directives={}, arguments={}, selections=[], alias=None)
                        ],
                    ),
                    SelectedField(
                        name="page",
                        directives={},
                        arguments={},
                        selections=[
                            SelectedField(name="description", directives={}, arguments={}, selections=[], alias=None)
                        ],
                        alias=None,
                    ),
                ],
                alias=None,
            )
        ]
    )

    assert is_querying_page_data(info)



from pydantic import BaseModel

from orchestrator.domain.base import ProductBlockModel, SubscriptionModel


class DomainModelChanges(BaseModel):
    create_products: dict[str, type[SubscriptionModel]] = {}
    delete_products: set[str] = set()
    create_product_to_block_relations: dict[str, set[str]] = {}
    delete_product_to_block_relations: dict[str, set[str]] = {}
    create_product_blocks: dict[str, type[ProductBlockModel]] = {}
    delete_product_blocks: set[str] = set()
    create_product_block_relations: dict[str, set[str]] = {}
    delete_product_block_relations: dict[str, set[str]] = {}
    create_product_fixed_inputs: dict[str, set[str]] = {}
    update_product_fixed_inputs: dict[str, dict[str, str]] = {}
    delete_product_fixed_inputs: dict[str, set[str]] = {}
    create_resource_types: set[str] = set()
    rename_resource_types: dict[str, str] = {}
    update_block_resource_types: dict[str, dict[str, str]] = {}
    delete_resource_types: set[str] = set()
    create_resource_type_relations: dict[str, set[str]] = {}
    create_resource_type_instance_relations: dict[str, set[str]] = {}
    delete_resource_type_relations: dict[str, set[str]] = {}


class DuplicateError(Exception):
    pass


class ModelUpdates(BaseModel):
    fixed_inputs: dict[str, dict[str, str]] = {}
    resource_types: dict[str, str] = {}
    block_resource_types: dict[str, dict[str, str]] = {}



from orchestrator import OrchestratorCore
from orchestrator.cli.main import app as core_cli
from orchestrator.settings import AppSettings

app = OrchestratorCore(base_settings=AppSettings())
if __name__ == "__main__":
    core_cli()



# Copyright 2019-2020 SURF.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import functools
from collections.abc import AsyncGenerator
from contextlib import asynccontextmanager
from os import getenv
from typing import Any, Callable
from uuid import UUID

import redis.exceptions
from anyio import CancelScope, get_cancelled_exc_class
from redis import Redis
from redis.asyncio import Redis as AIORedis
from redis.asyncio.client import Pipeline, PubSub
from redis.asyncio.retry import Retry
from redis.backoff import EqualJitterBackoff
from structlog import get_logger

from orchestrator.services.subscriptions import _generate_etag
from orchestrator.settings import app_settings
from orchestrator.utils.json import PY_JSON_TYPES, json_dumps, json_loads

logger = get_logger(__name__)

cache = Redis.from_url(str(app_settings.CACHE_URI))

ONE_WEEK = 3600 * 24 * 7


def caching_models_enabled() -> bool:
    return getenv("AIOCACHE_DISABLE", "0") == "0" and app_settings.CACHE_DOMAIN_MODELS


def to_redis(subscription: dict[str, Any]) -> str | None:
    if caching_models_enabled():
        logger.info("Setting cache for subscription", subscription=subscription["subscription_id"])
        etag = _generate_etag(subscription)
        cache.set(f"orchestrator:domain:{subscription['subscription_id']}", json_dumps(subscription), ex=ONE_WEEK)
        cache.set(f"orchestrator:domain:etag:{subscription['subscription_id']}", etag, ex=ONE_WEEK)
        return etag

    logger.warning("Caching disabled, not caching subscription", subscription=subscription["subscription_id"])
    return None


def from_redis(subscription_id: UUID) -> tuple[PY_JSON_TYPES, str] | None:
    log = logger.bind(subscription_id=subscription_id)
    if caching_models_enabled():
        log.debug("Try to retrieve subscription from cache")
        obj = cache.get(f"orchestrator:domain:{subscription_id}")
        etag = cache.get(f"orchestrator:domain:etag:{subscription_id}")
        if obj and etag:
            log.info("Retrieved subscription from cache")
            return json_loads(obj), etag.decode("utf-8")
        log.info("Subscription not found in cache")
        return None
    log.warning("Caching disabled, not loading subscription")
    return None


def delete_from_redis(subscription_id: UUID) -> None:
    if caching_models_enabled():
        logger.info("Deleting subscription object from cache", subscription_id=subscription_id)
        cache.delete(f"orchestrator:domain:{subscription_id}")
        cache.delete(f"orchestrator:domain:etag:{subscription_id}")
    else:
        logger.warning("Caching disabled, not deleting subscription", subscription=subscription_id)


def default_get_subscription_id(data: Any) -> UUID:
    if hasattr(data, "subscription_id"):
        return data.subscription_id
    if isinstance(data, dict):
        return data["subscription_id"]
    return data


def delete_subscription_from_redis(
    extract_fn: Callable[..., UUID] = default_get_subscription_id
) -> Callable[[Callable[..., Any]], Callable[..., Any]]:
    def _delete_subscription_from_redis(func: Callable[..., Any]) -> Callable[..., Any]:
        @functools.wraps(func)
        async def wrapper(*args: tuple, **kwargs: dict[str, Any]) -> Any:
            data = await func(*args, **kwargs)
            key = extract_fn(data)
            delete_from_redis(key)
            return data

        return wrapper

    return _delete_subscription_from_redis


async def delete_keys_matching_pattern(_cache: AIORedis, pattern: str, chunksize: int = 5000) -> int:
    """Delete all keys matching the given pattern.

    Usage:
        >>> await delete_keys_matching_pattern("orchestrator:foo:*")  # doctest:+SKIP
    """
    deleted = 0

    async def fetch(_cursor: int = 0) -> tuple[int, list[bytes]]:
        return await _cache.scan(cursor=_cursor, match=pattern, count=chunksize)

    cursor, keys = await fetch()
    while keys or cursor != 0:
        if keys:
            deleted += await _cache.delete(*keys)
        cursor, keys = await fetch(_cursor=cursor)

    logger.debug("Deleted keys matching pattern", pattern=pattern, deleted=deleted)
    return deleted


class RedisBroadcast:
    """Small wrapper around redis.asyncio.Redis used by websocket broadcasting.

    Note:
        redis.asyncio.Redis.from_url() returns a client which maintains a ConnectionPool.
        This instance is thread-safe and does not create connections until needed.
        However, you cannot instantiate this in one asyncio event loop and use it in
        another event loop, as the created connections in the pool will then only
        be usable by the loop they were created in.
    """

    client: AIORedis

    def __init__(self, redis_url: str):
        self.client = AIORedis.from_url(
            redis_url,
            retry_on_error=[redis.exceptions.ConnectionError],
            retry_on_timeout=True,
            retry=Retry(EqualJitterBackoff(base=0.05), 2),
        )
        self.redis_url = redis_url

    @asynccontextmanager
    async def pipeline(self) -> AsyncGenerator[Pipeline, None]:
        """Context to prepare a pipeline object for issueing multiple commands, such as .publish().

        Automatically executes the pipeline afterwards (unless there was an exception).
        """
        async with self.client.pipeline() as pipe:
            yield pipe
            await pipe.execute()

    @asynccontextmanager
    async def subscriber(self, *channels: str) -> AsyncGenerator[PubSub, None]:
        """Context to subscribe to one or more channels.

        Automatically unsubscribes and releases the connection afterwards.
        """
        pubsub = self.client.pubsub(ignore_subscribe_messages=True)

        async def do_teardown() -> None:
            if not pubsub.subscribed:
                return

            await pubsub.unsubscribe(*channels)
            await pubsub.aclose()  # type: ignore[attr-defined]

        try:
            await pubsub.subscribe(*channels)
            yield pubsub
        except get_cancelled_exc_class():
            # https://anyio.readthedocs.io/en/3.x/cancellation.html#finalization
            with CancelScope(shield=True):
                await do_teardown()
            raise
        finally:
            await do_teardown()

    async def connect(self) -> None:
        # Execute a simple command to ensure we can establish a connection
        result = await self.client.ping()
        logger.debug("RedisBroadcast can connect to redis", ping_result=result)

    async def disconnect(self) -> None:
        logger.debug("Closing redis client")
        await self.client.aclose()  # type: ignore[attr-defined]
        logger.debug("Closed redis client")



import pytest
from pydantic import Field

from orchestrator.db import ProductBlockTable, db
from orchestrator.domain.base import ProductBlockModel
from orchestrator.types import SubscriptionLifecycle


@pytest.fixture
def test_product_block_with_union(test_product_sub_block_one, test_product_sub_block_two):
    SubBlockOneForTestInactive, SubBlockOneForTestProvisioning, SubBlockOneForTest = test_product_sub_block_one
    SubBlockTwoForTestInactive, SubBlockTwoForTestProvisioning, SubBlockTwoForTest = test_product_sub_block_two

    class ProductBlockWithUnionForTestInactive(ProductBlockModel, product_block_name="ProductBlockWithUnionForTest"):
        union_block: SubBlockOneForTestInactive | SubBlockTwoForTestInactive | None = None
        int_field: int | None = None
        str_field: str | None = None
        list_field: list[int] = Field(default_factory=list)

    class ProductBlockWithUnionForTestProvisioning(
        ProductBlockWithUnionForTestInactive, lifecycle=[SubscriptionLifecycle.PROVISIONING]
    ):
        union_block: SubBlockOneForTestProvisioning | SubBlockTwoForTestProvisioning
        int_field: int
        str_field: str | None = None
        list_field: list[int]

    class ProductBlockWithUnionForTest(
        ProductBlockWithUnionForTestProvisioning, lifecycle=[SubscriptionLifecycle.ACTIVE]
    ):
        union_block: SubBlockOneForTest | SubBlockTwoForTest
        int_field: int
        str_field: str
        list_field: list[int]

    return ProductBlockWithUnionForTestInactive, ProductBlockWithUnionForTestProvisioning, ProductBlockWithUnionForTest


@pytest.fixture
def test_product_block_with_union_db(
    test_product_block_one_db,
    test_product_sub_block_one_db,
    test_product_sub_block_two_db,
    resource_type_int,
    resource_type_str,
    resource_type_list,
):
    product_block, product_sub_block = test_product_block_one_db
    product_union_sub_block = ProductBlockTable(
        name="ProductBlockWithUnionForTest", description="Test Union Sub Block", tag="TEST", status="active"
    )
    product_union_sub_block.resource_types = [resource_type_int, resource_type_str, resource_type_list]
    product_union_sub_block.depends_on.append(test_product_sub_block_one_db)
    product_union_sub_block.depends_on.append(test_product_sub_block_two_db)
    db.session.add(product_union_sub_block)
    product_block.depends_on.append(product_union_sub_block)
    db.session.commit()

    return product_block, product_sub_block, product_union_sub_block



from collections.abc import Generator
from datetime import datetime, timedelta
from typing import Any
from uuid import uuid4

import pytest
import pytz
from sqlalchemy import delete

from orchestrator.config.assignee import Assignee
from orchestrator.db import ProcessStepTable, ProcessSubscriptionTable, ProcessTable, db
from orchestrator.workflow import done, init, inputstep, step, workflow
from pydantic_forms.core import FormPage
from pydantic_forms.types import FormGenerator, UUIDstr
from pydantic_forms.validators import Choice
from test.unit_tests.workflows import WorkflowInstanceForTests


@pytest.fixture
def test_workflow(generic_subscription_1: UUIDstr, generic_product_type_1) -> Generator:
    _, GenericProductOne = generic_product_type_1

    @step("Insert UUID in state")
    def insert_object():
        return {"subscription_id": str(uuid4()), "model": GenericProductOne.from_subscription(generic_subscription_1)}

    @step("Test that it is a string now")
    def check_object(subscription_id: Any, model: dict) -> None:
        # This is actually a test. It would be nicer to have this in a proper test but that takes to much setup that
        # already happens here. So we hijack this fixture and run this test for all tests that use this fixture
        # (which should not be an issue)
        assert isinstance(subscription_id, str)
        assert isinstance(model, dict)

    @inputstep("Modify", assignee=Assignee.CHANGES)
    def modify(subscription_id: UUIDstr) -> FormGenerator:
        class TestChoice(Choice):
            A = "A"
            B = "B"
            C = "C"

        class TestForm(FormPage):
            generic_select: TestChoice

        user_input = yield TestForm
        return user_input.model_dump()

    @workflow("Workflow")
    def workflow_for_testing_processes_py():
        return init >> insert_object >> check_object >> modify >> done

    with WorkflowInstanceForTests(workflow_for_testing_processes_py, "workflow_for_testing_processes_py") as wf:
        yield wf


@pytest.fixture
def mocked_processes(test_workflow, generic_subscription_1, generic_subscription_2):
    first_datetime = datetime(2020, 1, 14, 9, 30, tzinfo=pytz.utc)

    def mock_process(subscription_id, status, started, assignee=Assignee.SYSTEM, is_task=False):
        process_id = uuid4()
        process = ProcessTable(
            process_id=process_id,
            workflow_id=test_workflow.workflow_id,
            last_status=status,
            last_step="Modify",
            started_at=started,
            last_modified_at=started + timedelta(minutes=10),
            assignee=assignee,
            is_task=is_task,
        )

        init_step = ProcessStepTable(process_id=process_id, name="Start", status="success", state={})
        db.session.add(process)

        if subscription_id:
            insert_step = ProcessStepTable(
                process_id=process_id,
                name="Insert UUID in state",
                status="success",
                state={"subscription_id": subscription_id},
            )
            check_step = ProcessStepTable(
                process_id=process_id,
                name="Test that it is a string now",
                status="success",
                state={"subscription_id": subscription_id},
            )
            step = ProcessStepTable(
                process_id=process_id, name="Modify", status="suspend", state={"subscription_id": subscription_id}
            )

            process_subscription = ProcessSubscriptionTable(process_id=process_id, subscription_id=subscription_id)

            db.session.add(init_step)
            db.session.add(insert_step)
            db.session.add(check_step)
            db.session.add(step)
            db.session.add(process_subscription)
        db.session.commit()

        return process_id

    return [
        mock_process(generic_subscription_1, "completed", first_datetime),
        mock_process(generic_subscription_1, "suspended", first_datetime + timedelta(days=1), assignee="NOC"),
        mock_process(generic_subscription_2, "completed", first_datetime + timedelta(days=2)),
        mock_process(generic_subscription_2, "failed", first_datetime + timedelta(days=3)),
        mock_process(generic_subscription_1, "resumed", first_datetime + timedelta(days=5)),
        mock_process(generic_subscription_1, "suspended", first_datetime + timedelta(days=1), is_task=True),
        mock_process(generic_subscription_2, "completed", first_datetime + timedelta(days=2), is_task=True),
        mock_process(None, "running", first_datetime + timedelta(days=4), is_task=True),
        mock_process(None, "resumed", first_datetime + timedelta(days=5), is_task=True),
    ]


@pytest.fixture
def mocked_processes_resumeall(test_workflow, generic_subscription_1, generic_subscription_2):
    first_datetime = datetime(2020, 2, 14, 9, 30, tzinfo=pytz.utc)

    def mock_process(subscription_id, status, started, assignee=Assignee.SYSTEM, is_task=False):
        process_id = uuid4()
        process = ProcessTable(
            process_id=process_id,
            workflow_id=test_workflow.workflow_id,
            last_status=status,
            last_step="Modify",
            started_at=started,
            last_modified_at=started + timedelta(minutes=10),
            assignee=assignee,
            is_task=is_task,
        )

        init_step = ProcessStepTable(process_id=process_id, name="Start", status="success", state={})
        db.session.add(process)

        if subscription_id:
            insert_step = ProcessStepTable(
                process_id=process_id,
                name="Insert UUID in state",
                status="success",
                state={"subscription_id": subscription_id},
            )
            check_step = ProcessStepTable(
                process_id=process_id,
                name="Test that it is a string now",
                status="success",
                state={"subscription_id": subscription_id},
            )
            step = ProcessStepTable(
                process_id=process_id, name="Modify", status="suspend", state={"subscription_id": subscription_id}
            )

            process_subscription = ProcessSubscriptionTable(process_id=process_id, subscription_id=subscription_id)

            db.session.add(init_step)
            db.session.add(insert_step)
            db.session.add(check_step)
            db.session.add(step)
            db.session.add(process_subscription)
        db.session.commit()

        return process_id

    mock_processes = [
        mock_process(generic_subscription_1, "api_unavailable", first_datetime, is_task=True),
        mock_process(
            generic_subscription_1, "suspended", first_datetime + timedelta(days=1), assignee="NOC", is_task=True
        ),
        mock_process(generic_subscription_2, "completed", first_datetime + timedelta(days=2), is_task=True),
        mock_process(generic_subscription_2, "running", first_datetime + timedelta(days=2), is_task=True),
        mock_process(generic_subscription_2, "failed", first_datetime + timedelta(days=3), is_task=True),
        mock_process(generic_subscription_1, "resumed", first_datetime + timedelta(days=5), is_task=True),
        mock_process(generic_subscription_1, "inconsistent_data", first_datetime + timedelta(days=1), is_task=True),
        mock_process(generic_subscription_2, "failed", first_datetime + timedelta(days=2), is_task=False),
        mock_process(None, "running", first_datetime + timedelta(days=4), is_task=True),
        mock_process(None, "resumed", first_datetime + timedelta(days=5), is_task=True),
    ]
    yield mock_processes
    db.session.execute(delete(ProcessTable).where(ProcessTable.process_id.in_(mock_processes)))



from collections.abc import Iterable
from itertools import groupby

import structlog
from sqlalchemy.dialects import postgresql
from sqlalchemy.sql.dml import UpdateBase

from orchestrator.domain.base import ProductBlockModel

logger = structlog.get_logger(__name__)


def sql_compile(sql: UpdateBase) -> str:
    sql_string = str(sql.compile(dialect=postgresql.dialect(), compile_kwargs={"literal_binds": True}))  # type: ignore[no-untyped-call]
    sql_string = sql_string.replace("\n", "")
    logger.debug("generated SQL", sql_string=sql_string)
    return sql_string


def generic_mapper(prop_name: str, model_diffs: dict[str, dict[str, set[str]]]) -> dict[str, set[str]]:
    model_by_resource = [
        (prop_name_value, model_name)
        for model_name, diff in model_diffs.items()
        for prop_name_value in diff.get(prop_name, [])
    ]
    grouped = groupby(model_by_resource, lambda x: x[0])
    return {k: {val[1] for val in v} for k, v in grouped}


def map_create_fixed_inputs(model_diffs: dict[str, dict[str, set[str]]]) -> dict[str, set[str]]:
    return generic_mapper("missing_fixed_inputs_in_db", model_diffs)


def map_delete_fixed_inputs(model_diffs: dict[str, dict[str, set[str]]]) -> dict[str, set[str]]:
    return generic_mapper("missing_fixed_inputs_in_model", model_diffs)


def map_create_resource_type_relations(model_diffs: dict[str, dict[str, set[str]]]) -> dict[str, set[str]]:
    return generic_mapper("missing_resource_types_in_db", model_diffs)


def map_delete_resource_type_relations(model_diffs: dict[str, dict[str, set[str]]]) -> dict[str, set[str]]:
    return generic_mapper("missing_resource_types_in_model", model_diffs)


def map_create_product_block_relations(model_diffs: dict[str, dict[str, set[str]]]) -> dict[str, set[str]]:
    return generic_mapper("missing_product_blocks_in_db", model_diffs)


def map_delete_product_block_relations(model_diffs: dict[str, dict[str, set[str]]]) -> dict[str, set[str]]:
    return generic_mapper("missing_product_blocks_in_model", model_diffs)


def get_product_block_names(pbs: list[type[ProductBlockModel]]) -> Iterable[str]:
    return filter(None, (pb.name for pb in pbs))



from unittest import mock

from sqlalchemy.orm import Relationship

from orchestrator.db.models import WorkflowTable
from orchestrator.graphql.utils.get_query_loaders import get_query_loaders
from test.unit_tests.graphql.utils.fixtures import LARGE_QUERY_SELECTED_FIELDS, SMALL_QUERY_SELECTED_FIELDS


def to_string(loader) -> list[str]:
    # Make loader human readable by str-formatting the Model relations that it joins
    return [str(elem) for elem in loader.path.path if isinstance(elem, Relationship)]


def test_get_query_loaders():
    # given
    mock_info = mock.Mock()
    mock_info.selected_fields = LARGE_QUERY_SELECTED_FIELDS

    # when
    query_loaders = get_query_loaders(mock_info, WorkflowTable)

    # then
    actual_loaders = [to_string(loader) for loader in query_loaders]
    expected_loaders = [
        [
            "WorkflowTable.products",
            "ProductTable.product_blocks",
            "ProductBlockTable.resource_types",
            "ResourceTypeTable.product_blocks",
        ],
        [
            "WorkflowTable.products",
            "ProductTable.product_blocks",
            "ProductBlockTable.in_use_by_block_relations",
            "ProductBlockRelationTable.in_use_by",
        ],
    ]

    assert sorted(actual_loaders) == sorted(expected_loaders)


def test_get_query_loaders_noop():
    # given
    mock_info = mock.Mock()
    mock_info.selected_fields = SMALL_QUERY_SELECTED_FIELDS

    # when
    query_loaders = get_query_loaders(mock_info, WorkflowTable)

    # then
    actual_loaders = [to_string(loader) for loader in query_loaders]
    expected_loaders = []

    assert sorted(actual_loaders) == sorted(expected_loaders)



from functools import reduce
from typing import Any, Callable, Iterator, NamedTuple, cast

import structlog
from sqlalchemy import inspect
from sqlalchemy.ext.associationproxy import ObjectAssociationProxyInstance
from sqlalchemy.orm import (
    InstrumentedAttribute,
    Load,
    RelationshipDirection,
    RelationshipProperty,
    joinedload,
    selectinload,
    subqueryload,
)
from sqlalchemy.orm.strategies import SubqueryLoader

from orchestrator.db import ALL_DB_MODELS
from orchestrator.db.database import BaseModel as DbBaseModel

logger = structlog.get_logger()

LoaderFunc = Callable[..., Load]


class AttrLoader(NamedTuple):
    loader_fn: LoaderFunc
    attr: InstrumentedAttribute
    next_model: type[DbBaseModel]


_MODEL_LOADERS: dict[type[DbBaseModel], dict[str, list[AttrLoader]]] = {}


def _relation_type_to_loader_func(relationship: RelationshipProperty) -> LoaderFunc:
    match relationship.direction:
        case RelationshipDirection.MANYTOONE:
            # Always eagerload this through a join
            return cast(LoaderFunc, joinedload)
        case RelationshipDirection.MANYTOMANY | RelationshipDirection.ONETOMANY:
            # Respect the strategy configured on the relationship
            loader_fn = subqueryload if isinstance(relationship.strategy, SubqueryLoader) else selectinload
            return cast(LoaderFunc, loader_fn)
    raise TypeError(f"Unrecognized relationship direction {relationship.direction}")


def _inspect_relationships(model: type[DbBaseModel]) -> Iterator[tuple[str, list[AttrLoader]]]:
    model_inspect: Any = inspect(model)

    def make_attr_loader(model: type[DbBaseModel], relation: RelationshipProperty) -> AttrLoader:
        loader = _relation_type_to_loader_func(relation)
        attr = cast(InstrumentedAttribute, getattr(model, relation.key))
        next_model = cast(type[DbBaseModel], relation.entity.entity)
        return AttrLoader(loader, attr, next_model)

    yield from ((relation.key, [make_attr_loader(model, relation)]) for relation in model_inspect.relationships)


def _inspect_assocation_proxies(first_model: type[DbBaseModel]) -> Iterator[tuple[str, list[AttrLoader]]]:

    # Association proxies are not inspectable so they require some extra work
    # https://github.com/sqlalchemy/sqlalchemy/discussions/10047
    # E.g. for the association_proxy ProductBlockTable.in_use_by we want to end up with the loader:
    #   selectinload(ProductBlockTable.in_use_by_block_relations).joinedload(ProductBlockRelationTable.in_use_by)
    def is_assoc_proxy(attr_name: str) -> bool:
        if attr_name.startswith("_"):
            return False
        attr = getattr(first_model, attr_name)
        return isinstance(attr, ObjectAssociationProxyInstance)

    assoc_proxy_attrs = (attr_name for attr_name in dir(first_model) if is_assoc_proxy(attr_name))

    def to_relation_names(attr_name: str) -> tuple[str, str]:
        attr: ObjectAssociationProxyInstance = getattr(first_model, attr_name)
        return attr.target_collection, attr.value_attr

    def make_attr_loader(model: type[DbBaseModel], relation_name: str) -> AttrLoader:
        inspected: Any = inspect(model)
        relation: RelationshipProperty = inspected.relationships[relation_name]
        loader = _relation_type_to_loader_func(relation)
        attr = cast(InstrumentedAttribute, getattr(model, relation_name))
        return AttrLoader(loader, attr, model)

    model_inspect: Any = inspect(first_model)
    for attr_name in assoc_proxy_attrs:
        first_relation_name, second_relation_name = to_relation_names(attr_name)
        first_attr_loader = make_attr_loader(first_model, first_relation_name)

        first_relation: RelationshipProperty = model_inspect.relationships[first_relation_name]
        second_model = cast(type[DbBaseModel], first_relation.entity.entity)
        second_attr_loader = make_attr_loader(second_model, second_relation_name)

        yield attr_name, [first_attr_loader, second_attr_loader]


def _inspect_model(model: type[DbBaseModel]) -> Iterator[tuple[str, list[AttrLoader]]]:
    yield from _inspect_relationships(model)
    yield from _inspect_assocation_proxies(model)


def init_model_loaders() -> None:
    """Inspects relationships from all SQLAlchemy models to prepare loader functions.

    Is called once during startup of the application.

    As an example, an excerpt from one of the biggest models: ProductBlockTable
        <class 'orchestrator.db.models.ProductBlockTable'>: {
          'products': [
            AttrLoader(
              loader_fn=<function selectinload at 0x...>,
              attr=<sqlalchemy.orm.attributes.InstrumentedAttribute object at 0x...>,
              next_model=<class 'orchestrator.db.models.ProductTable'>
            )
          ],
          'in_use_by': [
            AttrLoader(
              loader_fn=<function selectinload at 0x...>,
              attr=<sqlalchemy.orm.attributes.InstrumentedAttribute object at 0x...>,
              next_model=<class 'orchestrator.db.models.ProductBlockTable'>
            ),
            AttrLoader(
              loader_fn=<function joinedload at 0x...>,
              attr=<sqlalchemy.orm.attributes.InstrumentedAttribute object at 0x...>,
              next_model=<class 'orchestrator.db.models.ProductBlockRelationTable'>
            )
          ],
          ...
        }
    """
    for model in ALL_DB_MODELS:
        _MODEL_LOADERS[model] = dict(_inspect_model(model))


def lookup_attr_loaders(model: type[DbBaseModel], attr: str) -> list[AttrLoader]:
    """Return loader(s) for an attribute on the given model."""
    return _MODEL_LOADERS.get(model, {}).get(attr, [])


def join_attr_loaders(loaders: list[AttrLoader]) -> Load | None:
    """Given 1 or more attribute loaders, instantiate and chain them together."""
    if not loaders:
        return None

    first_loader, *other_loaders = loaders

    loader_fn = first_loader.loader_fn(first_loader.attr)
    if not other_loaders:
        return loader_fn

    def chain_loader_func(final_loader: Load, next: AttrLoader) -> Load:
        return getattr(final_loader, next.loader_fn.__name__)(next.attr)

    return reduce(chain_loader_func, other_loaders, loader_fn)



from typing import Annotated

import structlog
from orchestrator.domain import SubscriptionModel
from orchestrator.forms import FormPage
from orchestrator.forms.validators import CustomerId, Divider
from orchestrator.types import FormGenerator, State, SubscriptionLifecycle, UUIDstr
from orchestrator.workflow import StepList, begin, step
from orchestrator.workflows.steps import set_status
from orchestrator.workflows.utils import modify_workflow
from pydantic import AfterValidator
from pydantic_forms.validators import ReadOnlyField

from products.product_blocks.example1 import ExampleStrEnum1
from products.product_types.example1 import Example1, Example1Provisioning
from workflows.example1.shared.forms import must_be_unused_to_change_mode_validator
from workflows.shared import modify_summary_form


def subscription_description(subscription: SubscriptionModel) -> str:
    """The suggested pattern is to implement a subscription service that generates a subscription specific
    description, in case that is not present the description will just be set to the product name.
    """
    return f"{subscription.product.name} subscription"


logger = structlog.get_logger(__name__)

validated_example_str_enum_1 = Annotated[ExampleStrEnum1, AfterValidator(must_be_unused_to_change_mode_validator)]


def initial_input_form_generator(subscription_id: UUIDstr) -> FormGenerator:
    subscription = Example1.from_subscription(subscription_id)
    example1 = subscription.example1

    # TODO fill in additional fields if needed

    class ModifyExample1Form(FormPage):
        customer_id: CustomerId = subscription.customer_id  # type: ignore

        divider_1: Divider

        unmodifiable_str: ReadOnlyField(example1.unmodifiable_str)
        annotated_int: ReadOnlyField(example1.annotated_int)
        example_str_enum_1: validated_example_str_enum_1 = example1.example_str_enum_1
        modifiable_boolean: bool = example1.modifiable_boolean
        always_optional_str: str | None = example1.always_optional_str

    user_input = yield ModifyExample1Form
    user_input_dict = user_input.dict()

    summary_fields = [
        "example_str_enum_1",
        "unmodifiable_str",
        "modifiable_boolean",
        "annotated_int",
        "always_optional_str",
    ]
    yield from modify_summary_form(user_input_dict, subscription.example1, summary_fields)

    return user_input_dict | {"subscription": subscription}


@step("Update subscription")
def update_subscription(
    subscription: Example1Provisioning,
    example_str_enum_1: ExampleStrEnum1,
    modifiable_boolean: bool,
    always_optional_str: str | None,
) -> State:
    # TODO: get all modified fields
    subscription.example1.example_str_enum_1 = example_str_enum_1
    subscription.example1.modifiable_boolean = modifiable_boolean
    subscription.example1.always_optional_str = always_optional_str

    return {"subscription": subscription}


@step("Update subscription description")
def update_subscription_description(subscription: Example1) -> State:
    subscription.description = subscription_description(subscription)
    return {"subscription": subscription}


additional_steps = begin


@modify_workflow("Modify example1", initial_input_form=initial_input_form_generator, additional_steps=additional_steps)
def modify_example1() -> StepList:
    return (
        begin
        >> set_status(SubscriptionLifecycle.PROVISIONING)
        >> update_subscription
        >> update_subscription_description
        # TODO add additional steps if needed
        >> set_status(SubscriptionLifecycle.ACTIVE)
    )



from dirty_equals import IsFloat
from sqlalchemy import text

from orchestrator.db import db
from orchestrator.db.listeners import disable_listeners, monitor_sqlalchemy_queries


def test_monitor_sqlalchemy_queries():
    monitor_sqlalchemy_queries()

    try:

        db.session.execute(text("select 1"))

        stats = db.session.connection().info.copy()
        assert stats == {
            "queries_completed": 1,
            "queries_started": 1,
            "query_start_time": [],
            "query_time_spent": IsFloat,
        }
    finally:
        disable_listeners()



from orchestrator.utils.get_updated_properties import get_updated_properties


def test_get_updated_fields_with_root_props():
    obj1 = {
        "unchanged": "unchanged",
        "update": "to_update",
        "removed": "to_remove",
    }

    obj2 = {
        "unchanged": "unchanged",
        "update": "updated_prop",
        "new_prop": "new",
    }

    updated_props = get_updated_properties(obj1, obj2)

    assert updated_props == {
        "update": "updated_prop",
        "new_prop": "new",
    }


def test_get_updated_fields_with_list_props():
    obj1 = {
        "list_unchanged": [1, 2, 3],
        "list_update_item": [1, 2, 3],
        "list_update_and_new": [1, 2, 3],
        "list_update_and_new_and_remove": [1, 2, 3],
        "list_remove": [1, 2, 3],
    }

    obj2 = {
        "list_unchanged": [1, 2, 3],
        "list_update_item": [80, 2, 20],
        "list_update_and_new": [80, 2, 20, 30],
        "list_update_and_new_and_remove": [80, 2],
        "list_new": [10, 11, 12],
    }

    updated_props = get_updated_properties(obj1, obj2)

    assert updated_props == {
        "list_update_item": [80, 2, 20],
        "list_update_and_new": [80, 2, 20, 30],
        "list_update_and_new_and_remove": [80, 2],
        "list_new": [10, 11, 12],
    }


def test_get_updated_fields_with_dict_props():
    obj1 = {
        "dict_unchanged": {"unchanged": "unchanged"},
        "dict_update": {
            "unchanged": "unchanged",
            "update": "to update",
        },
        "dict_update_and_new": {
            "unchanged": "unchanged",
            "update": "to update",
        },
        "dict_update_and_remove": {
            "unchanged": "unchanged",
            "update": "to update",
            "update_dict": {"nested": "dict"},
            "remove": "to remove",
        },
        "dict_remove": {"remove": "to remove"},
    }

    obj2 = {
        "dict_unchanged": {"unchanged": "unchanged"},
        "dict_update": {
            "unchanged": "unchanged",
            "update": "updated prop",
        },
        "dict_update_and_new": {
            "unchanged": "unchanged",
            "update": "updated prop",
            "new_dict": {"nested": "dict"},
        },
        "dict_update_and_remove": {
            "unchanged": "unchanged",
            "update": "updated prop",
            "update_dict": {"nested": "dict update"},
        },
        "dict_new": {
            "new": "dict",
        },
    }

    updated_props = get_updated_properties(obj1, obj2)

    assert updated_props == {
        "dict_update": {"update": "updated prop"},
        "dict_update_and_new": {"update": "updated prop", "new_dict": {"nested": "dict"}},
        "dict_update_and_remove": {"update": "updated prop", "update_dict": {"nested": "dict update"}},
        "dict_new": {"new": "dict"},
    }


def test_get_updated_fields_with_mixed_props():
    obj1 = {
        "test": "test",
        "update": "test",
        "nested": {"key1": "value1", "key2": "value2"},
        "list": [1, 2, 3],
        "list_dict": [{"test": "test"}],
    }

    obj2 = {
        "test": "test",
        "update": "updated",
        "nested": {"key1": "value1", "key2": "updated_value"},
        "list": [1, 2, 4],
        "list_dict": [{"test": "update"}],
    }

    updated_props = get_updated_properties(obj1, obj2)

    assert updated_props == {
        "update": "updated",
        "nested": {"key2": "updated_value"},
        "list": [1, 2, 4],
        "list_dict": [{"test": "update"}],
    }



# Copyright 2019-2020 SURF.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


def pre_mutation(context):
    """Prepare mutmut context.

    This helps skipping needless mutations.
    """
    line = context.current_source_line.strip()
    if context.current_line_index != 0:
        prev_line = context.source_by_line_number[context.current_line_index - 1].strip()
    else:
        prev_line = ""

    if line.startswith("logger.") or prev_line.startswith("logger."):
        context.skip = True
    if line.startswith("logger = structlog"):
        context.skip = True
    if line.startswith("cls.__doc__"):
        context.skip = True

    # This file is copied verbatim and is not tested
    if context.filename.endswith("crypt.py"):
        context.skip = True



from uuid import uuid4

import pytest
from pydantic import ValidationError

from orchestrator.db import db
from orchestrator.types import SubscriptionLifecycle
from test.unit_tests.fixtures.products.product_blocks.product_block_one import DummyEnum


def test_product_model_with_union_type_directly_below(
    test_union_product,
    test_union_type_product,
    test_product_sub_block_one,
    test_product_block_one,
    sub_one_subscription_1,
):
    UnionProductInactive, _, UnionProduct = test_union_type_product
    SubBlockOneForTestInactive, _, SubBlockOneForTest = test_product_sub_block_one
    ProductBlockOneForTestInactive, _, _ = test_product_block_one

    union_subscription_inactive = UnionProductInactive.from_product_id(
        product_id=test_union_product, customer_id=str(uuid4())
    )

    union_subscription_inactive.test_block = ProductBlockOneForTestInactive.new(
        subscription_id=union_subscription_inactive.subscription_id,
        int_field=3,
        str_field="",
        list_field=[1],
        enum_field=DummyEnum.FOO,
        sub_block=SubBlockOneForTestInactive.new(
            subscription_id=union_subscription_inactive.subscription_id, int_field=3, str_field=""
        ),
        sub_block_2=SubBlockOneForTestInactive.new(
            subscription_id=union_subscription_inactive.subscription_id, int_field=3, str_field=""
        ),
    )

    with pytest.raises(ValidationError) as error:
        UnionProduct.from_other_lifecycle(union_subscription_inactive, SubscriptionLifecycle.ACTIVE)
    assert error.value.errors()[0]["msg"] == "Input should be a valid dictionary or instance of ProductBlockOneForTest"

    new_sub_block = SubBlockOneForTest.new(
        subscription_id=union_subscription_inactive.subscription_id, int_field=1, str_field="2"
    )
    union_subscription_inactive.union_block = new_sub_block
    union_subscription_inactive.save()

    assert union_subscription_inactive.diff_product_in_database(union_subscription_inactive.product.product_id) == {}
    union_subscription = UnionProduct.from_other_lifecycle(union_subscription_inactive, SubscriptionLifecycle.ACTIVE)

    union_subscription.union_block = sub_one_subscription_1.test_block

    with pytest.raises(ValueError) as exc:
        union_subscription.save()
        assert (
            str(exc)
            == "Attempting to save a Foreign `Subscription Instance` directly below a subscription. This is not allowed."
        )


def test_union_product_block_as_sub(
    test_union_sub_product,
    test_union_type_sub_product,
    test_product_block_with_union,
    test_product_type_sub_one,
    sub_one_subscription_1,
):
    UnionProductSubInactive, _, UnionProductSub = test_union_type_sub_product
    UnionProductBlockForTestInactive, _, _ = test_product_block_with_union
    _, _, ProductSubOne = test_product_type_sub_one

    union_subscription_inactive = UnionProductSubInactive.from_product_id(
        product_id=test_union_sub_product, customer_id=str(uuid4())
    )
    union_subscription_inactive.test_block = UnionProductBlockForTestInactive.new(
        subscription_id=union_subscription_inactive.subscription_id
    )
    union_subscription_inactive.save()

    union_subscription_inactive.test_block.int_field = 1
    union_subscription_inactive.test_block.str_field = "blah"
    union_subscription_inactive.test_block.union_block = sub_one_subscription_1.test_block

    union_subscription_inactive.test_block.list_field = [2]

    union_subscription = UnionProductSub.from_other_lifecycle(
        union_subscription_inactive, status=SubscriptionLifecycle.ACTIVE
    )
    union_subscription.save()

    # This needs to happen in the test due to the fact it is using cached objects.
    db.session.commit()
    assert union_subscription.diff_product_in_database(test_union_sub_product) == {}

    union_subscription_from_database = UnionProductSub.from_subscription(union_subscription.subscription_id)

    assert type(union_subscription_from_database) is type(union_subscription)
    assert union_subscription_from_database.test_block.int_field == union_subscription.test_block.int_field
    assert union_subscription_from_database.test_block.str_field == union_subscription.test_block.str_field
    assert (
        union_subscription_from_database.test_block.union_block.subscription_instance_id
        == sub_one_subscription_1.test_block.subscription_instance_id
    )
    # Do not allow subscriptions that are in use by other subscriptions make an unsafe transition.
    with pytest.raises(ValueError):
        ProductSubOne.from_other_lifecycle(sub_one_subscription_1, SubscriptionLifecycle.TERMINATED)



from collections.abc import Generator
from itertools import count
from typing import TYPE_CHECKING, Annotated, Any
from uuid import UUID

import strawberry
from pydantic.alias_generators import to_camel as to_lower_camel
from strawberry.scalars import JSON

from orchestrator.graphql.schemas.product_block import owner_subscription_resolver
from orchestrator.utils.get_subscription_dict import get_subscription_dict

if TYPE_CHECKING:
    from orchestrator.graphql.schemas.subscription import SubscriptionInterface


@strawberry.type
class ProductBlockInstance:
    id: int
    parent: int | None
    subscription_instance_id: UUID
    owner_subscription_id: UUID
    in_use_by_relations: list[JSON]
    product_block_instance_values: JSON
    subscription: (
        Annotated["SubscriptionInterface", strawberry.lazy("orchestrator.graphql.schemas.subscription")] | None
    ) = strawberry.field(
        description="resolve to subscription of the product block", resolver=owner_subscription_resolver
    )


def is_product_block(candidate: Any) -> bool:
    if isinstance(candidate, dict):
        # TODO: also filter on tag (needs addition of tag in orchestrator endpoint)
        # NOTE: this crosses subscription boundaries. If needed we can add an additional filter to limit that.
        return candidate.get("owner_subscription_id", None)
    return False


def get_all_product_blocks(subscription: dict[str, Any], _tags: list[str] | None) -> list[dict[str, Any]]:
    gen_id = count()

    def locate_product_block(candidate: dict[str, Any]) -> Generator:
        def new_product_block(item: dict[str, Any]) -> Generator:
            enriched_item = item | {"id": next(gen_id), "parent": candidate.get("id")}
            yield enriched_item
            yield from locate_product_block(enriched_item)

        for value in candidate.values():
            if is_product_block(value):
                yield from new_product_block(value)
            elif isinstance(value, list):
                for item in value:
                    if is_product_block(item):
                        yield from new_product_block(item)

    return list(locate_product_block(subscription))


pb_instance_property_keys = ("id", "parent", "owner_subscription_id", "subscription_instance_id", "in_use_by_relations")


async def get_subscription_product_blocks(
    subscription_id: UUID, tags: list[str] | None = None, product_block_instance_values: list[str] | None = None
) -> list[ProductBlockInstance]:
    subscription, _ = await get_subscription_dict(subscription_id)

    def to_product_block(product_block: dict[str, Any]) -> ProductBlockInstance:
        def is_resource_type(candidate: Any) -> bool:
            return not isinstance(candidate, (list, dict))

        def requested_resource_type(key: str) -> bool:
            return not product_block_instance_values or key in product_block_instance_values

        def included(key: str, value: Any) -> bool:
            return is_resource_type(value) and requested_resource_type(key) and key not in pb_instance_property_keys

        return ProductBlockInstance(
            id=product_block["id"],
            parent=product_block.get("parent"),
            owner_subscription_id=product_block["owner_subscription_id"],
            subscription_instance_id=product_block["subscription_instance_id"],
            product_block_instance_values=[
                {"field": to_lower_camel(k), "value": v if isinstance(v, (str, int, float, type(None))) else str(v)}
                for k, v in product_block.items()
                if included(k, v)
            ],
            in_use_by_relations=product_block.get("in_use_by_relations", []),
        )

    product_blocks = (to_product_block(product_block) for product_block in get_all_product_blocks(subscription, tags))
    return [product_block for product_block in product_blocks if product_block.product_block_instance_values]



from orchestrator.domain.base import SubscriptionModel
from orchestrator.types import SubscriptionLifecycle

from products.product_blocks.example4 import Example4Block, Example4BlockInactive, Example4BlockProvisioning


class Example4Inactive(SubscriptionModel, is_base=True):
    example4: Example4BlockInactive


class Example4Provisioning(Example4Inactive, lifecycle=[SubscriptionLifecycle.PROVISIONING]):
    example4: Example4BlockProvisioning


class Example4(Example4Provisioning, lifecycle=[SubscriptionLifecycle.ACTIVE]):
    example4: Example4Block



# Copyright 2019-2020 SURF.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from collections.abc import Callable
from http import HTTPStatus
from typing import Any
from uuid import UUID

import structlog
from celery.result import AsyncResult
from kombu.exceptions import ConnectionError, OperationalError

from orchestrator import app_settings
from orchestrator.api.error_handling import raise_status
from orchestrator.db import ProcessTable, db
from orchestrator.services.processes import create_process, delete_process
from orchestrator.targets import Target
from orchestrator.types import State
from orchestrator.workflows import get_workflow

SYSTEM_USER = "SYSTEM"

logger = structlog.get_logger(__name__)


def _block_when_testing(task_result: AsyncResult) -> None:
    # Enables "Sync celery tasks. This will let the app wait until celery completes"
    if app_settings.TESTING:
        process_id = task_result.get()
        if not process_id:
            raise RuntimeError("Celery worker has failed to resume process")


def _celery_start_process(
    workflow_key: str, user_inputs: list[State] | None, user: str = SYSTEM_USER, **kwargs: Any
) -> UUID:
    """Client side call of Celery."""
    from orchestrator.services.tasks import NEW_TASK, NEW_WORKFLOW, get_celery_task

    workflow = get_workflow(workflow_key)
    if not workflow:
        raise_status(HTTPStatus.NOT_FOUND, "Workflow does not exist")

    task_name = NEW_TASK if workflow.target == Target.SYSTEM else NEW_WORKFLOW
    trigger_task = get_celery_task(task_name)
    pstat = create_process(workflow_key, user_inputs, user)
    tasks = pstat.state.s
    try:
        result = trigger_task.delay(pstat.process_id, workflow_key, tasks, user)
        _block_when_testing(result)
        return pstat.process_id
    except (ConnectionError, OperationalError) as e:
        # If connection to Redis fails and process can't be started, we need to remove the created process
        logger.warning("Connection error when submitting task to Celery. Delete newly created process from database.")
        delete_process(pstat.process_id)
        raise e


def _celery_resume_process(
    process: ProcessTable,
    *,
    user_inputs: list[State] | None,
    user: str | None,
    **kwargs: Any,
) -> UUID:
    """Client side call of Celery."""
    from orchestrator.services.processes import load_process
    from orchestrator.services.tasks import RESUME_TASK, RESUME_WORKFLOW, get_celery_task

    pstat = load_process(process)
    last_process_status = process.last_status
    workflow = pstat.workflow

    task_name = RESUME_TASK if workflow.target == Target.SYSTEM else RESUME_WORKFLOW
    trigger_task = get_celery_task(task_name)
    try:
        _celery_set_process_status_resumed(process)
        result = trigger_task.delay(pstat.process_id, user_inputs, user)
        _block_when_testing(result)

        return pstat.process_id
    except (ConnectionError, OperationalError) as e:
        logger.warning(
            "Connection error when submitting task to celery. Resetting process status back",
            current_status=process.last_status,
            last_status=last_process_status,
        )
        _celery_set_process_status(process, last_process_status)
        raise e


def _celery_set_process_status(process: ProcessTable, status: str) -> None:
    process.last_status = status
    db.session.add(process)
    db.session.commit()


def _celery_set_process_status_resumed(process: ProcessTable) -> None:
    """Set the process status to RESUMED to prevent re-adding to task queue.

    Args:
        process: Process from database
    """
    from orchestrator.db import db
    from orchestrator.workflow import ProcessStatus

    process.last_status = ProcessStatus.RESUMED
    db.session.add(process)
    db.session.commit()


def _celery_validate(validation_workflow: str, json: list[State] | None) -> None:
    _celery_start_process(validation_workflow, user_inputs=json)


CELERY_EXECUTION_CONTEXT: dict[str, Callable] = {
    "start": _celery_start_process,
    "resume": _celery_resume_process,
    "validate": _celery_validate,
}



import pathlib
from pathlib import Path
from unittest import mock

from typer.testing import CliRunner

from orchestrator.cli.generate import app
from orchestrator.domain.base import ProductBlockModel
from orchestrator.types import SubscriptionLifecycle


class MyExistingProductBlockInactive(ProductBlockModel, product_block_name="My Existing Product Block"):
    user_id: int | None = None


class MyExistingProductBlockProvisioning(
    MyExistingProductBlockInactive, lifecycle=[SubscriptionLifecycle.PROVISIONING]
):
    user_id: int


class MyExistingProductBlock(MyExistingProductBlockProvisioning, lifecycle=[SubscriptionLifecycle.ACTIVE]):
    pass


def read_file(path: str) -> str:
    file = Path(__file__).resolve().parent / "data" / path
    return str(file)


def test_generate_product():
    runner = CliRunner()
    result = runner.invoke(app, ["product", "--config-file", read_file("product_config3.yaml"), "--dryrun"])
    assert "class MyProductInactive" in result.stdout
    assert "class MyProductProvisioning" in result.stdout
    assert "class MyProduct" in result.stdout
    assert result.exit_code == 0


@mock.patch("orchestrator.cli.generator.generator.product_block.get_existing_product_blocks")
def test_generate_product_blocks(existing_product_blocks_mock):
    existing_product_blocks_mock.return_value = {
        "MyExistingProductBlock": "products.product_blocks.my_existing_product_block"
    }
    runner = CliRunner()
    result = runner.invoke(app, ["product-blocks", "--config-file", read_file("product_config3.yaml"), "--dryrun"])
    assert "class MyIntermediateProductBlockInactive" in result.stdout
    assert "class MyIntermediateProductBlockProvisioning" in result.stdout
    assert "class MyIntermediateProductBlock" in result.stdout
    assert result.exit_code == 0


def test_generate_workflows():
    runner = CliRunner()
    result = runner.invoke(app, ["workflows", "--config-file", read_file("product_config3.yaml"), "--dryrun"])
    assert "@create_workflow(" in result.stdout
    assert "@modify_workflow(" in result.stdout
    assert "@terminate_workflow(" in result.stdout
    assert "@validate_workflow(" in result.stdout
    assert result.exit_code == 0


@mock.patch.object(pathlib.Path, "mkdir")
def test_generate_unit_tests(mkdir_mock):
    runner = CliRunner()
    result = runner.invoke(app, ["unit-tests", "--config-file", read_file("product_config3.yaml"), "--dryrun"])
    assert "def test_" in result.stdout
    assert result.exit_code == 0



# Copyright 2022 SURF.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import json
from http import HTTPStatus
from uuid import uuid4


def build_simple_query(subscription_id):

    q = """query SubscriptionQuery($id: UUID!) {
          subscription(id: $id) {
            insync
            status
            }
        }"""
    return json.dumps(
        {
            "operationName": "SubscriptionQuery",
            "query": q,
            "variables": {
                "id": str(subscription_id),
            },
        }
    ).encode("utf-8")


def build_complex_query(subscription_id):
    q = """query SubscriptionQuery($id: UUID!) {
        subscription(id: $id) {
            insync
            __typename
            product {
                status
            }
            ... on ProductSubListUnionSubscription {
                testBlock {
                    intField
                }
            }
        }
    }"""
    return json.dumps(
        {
            "operationName": "SubscriptionQuery",
            "query": q,
            "variables": {
                "id": str(subscription_id),
            },
        }
    ).encode("utf-8")


def test_single_simple_subscription(fastapi_app_graphql, test_client, product_sub_list_union_subscription_1):
    test_query = build_simple_query(subscription_id=product_sub_list_union_subscription_1)
    response = test_client.post("/api/graphql", content=test_query, headers={"Content-Type": "application/json"})
    assert response.status_code == HTTPStatus.OK
    assert response.json() == {"data": {"subscription": {"insync": True, "status": "ACTIVE"}}}


def test_single_complex_subscription(
    fastapi_app_graphql, test_client, product_sub_list_union_subscription_1, test_product_type_sub_list_union
):
    _, _, ProductSubListUnion = test_product_type_sub_list_union
    test_query = build_complex_query(subscription_id=product_sub_list_union_subscription_1)
    response = test_client.post("/api/graphql", content=test_query, headers={"Content-Type": "application/json"})
    assert response.status_code == HTTPStatus.OK
    assert response.json() == {
        "data": {
            "subscription": {
                "__typename": "ProductSubListUnionSubscription",
                "insync": True,
                "product": {"status": "ACTIVE"},
                "testBlock": {"intField": 1},
            }
        }
    }


def test_subscription_does_not_exist(fastapi_app_graphql, test_client):
    test_query = build_simple_query(uuid4())
    response = test_client.post("/api/graphql", content=test_query, headers={"Content-Type": "application/json"})
    assert response.status_code == HTTPStatus.OK
    assert response.json() == {"data": {"subscription": None}}



from copy import deepcopy
from unittest import mock
from uuid import uuid4

import pytest
from pydantic import ValidationError

from orchestrator.forms.validators import ProductId, product_id
from pydantic_forms.core import FormPage
from test.unit_tests.helpers import URL_MISSING, URL_PARSING, URL_VALUE


@mock.patch("orchestrator.forms.validators.product_id.get_product_by_id")
def test_product_id(mock_get_product_by_id):
    product_x_id = uuid4()
    product_y_id = uuid4()

    mock_get_product_by_id.side_effect = [product_x_id, product_y_id]

    class Form(FormPage):
        product_x: product_id([product_x_id])
        product_id: ProductId

    data = Form(product_id=product_y_id, product_x=product_x_id)
    assert data.product_x == product_x_id
    assert data.product_id == product_y_id
    mock_get_product_by_id.assert_has_calls(
        [mock.call(product_x_id, join_fixed_inputs=False), mock.call(product_y_id, join_fixed_inputs=False)]
    )


def test_product_id_schema():
    product_x_id = uuid4()

    class Form(FormPage):
        product_x: product_id([product_x_id])
        product_id: ProductId

    assert Form.model_json_schema() == {
        "additionalProperties": False,
        "properties": {
            "product_id": {"format": "productId", "title": "Product Id", "type": "string"},
            "product_x": {
                "format": "productId",
                "title": "Product X",
                "type": "string",
                "uniforms": {"productIds": [str(product_x_id)]},
            },
        },
        "required": ["product_x", "product_id"],
        "title": "unknown",
        "type": "object",
    }


def stringify_exceptions(error_list):
    # without this the error lists are not considered equal
    list_copy = deepcopy(error_list)
    for error in list_copy:
        if "error" in error.get("ctx", {}):
            error["ctx"]["error"] = str(error["ctx"]["error"])
    return list_copy


@mock.patch("orchestrator.forms.validators.product_id.get_product_by_id")
def test_product_id_nok(mock_get_product_by_id):
    product_x_id = uuid4()
    product_y_id = uuid4()

    mock_get_product_by_id.side_effect = [product_y_id, None]

    class Form(FormPage):
        product_x: product_id([product_x_id])
        product_id: ProductId

    with pytest.raises(ValidationError) as error_info:
        Form(product_id=product_y_id, product_x=product_y_id)

    expected = [
        {
            "ctx": {"error": f"value is not a valid enumeration member; permitted: '{product_x_id}'"},
            "input": product_y_id,
            "loc": ("product_x",),
            "msg": f"Value error, value is not a valid enumeration member; permitted: '{product_x_id}'",
            "type": "value_error",
        }
        | URL_VALUE,
        {
            "type": "value_error",
            "loc": ("product_id",),
            "msg": "Value error, Product not found",
            "input": product_y_id,
            "ctx": {"error": ValueError("Product not found")},
        }
        | URL_VALUE,
    ]

    actual = error_info.value.errors()
    assert stringify_exceptions(expected) == stringify_exceptions(actual)
    mock_get_product_by_id.assert_has_calls(
        [mock.call(product_y_id, join_fixed_inputs=False), mock.call(product_y_id, join_fixed_inputs=False)]
    )

    with pytest.raises(ValidationError) as error_info:
        Form(product_id="INCOMPLETE")

    expected = [
        {"type": "missing", "loc": ("product_x",), "msg": "Field required", "input": {"product_id": "INCOMPLETE"}}
        | URL_MISSING,
        {
            "type": "uuid_parsing",
            "loc": ("product_id",),
            "msg": "Input should be a valid UUID, invalid character: expected an optional prefix of `urn:uuid:` followed by [0-9a-fA-F-], found `I` at 1",
            "input": "INCOMPLETE",
            "ctx": {
                "error": "invalid character: expected an optional prefix of `urn:uuid:` followed by [0-9a-fA-F-], found `I` at 1"
            },
        }
        | URL_PARSING,
    ]
    assert expected == error_info.value.errors()



# Copyright 2019-2020 SURF.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from http import HTTPStatus
from uuid import UUID

from fastapi.routing import APIRouter
from sqlalchemy import select
from sqlalchemy.orm import joinedload, selectinload

from orchestrator.api.error_handling import raise_status
from orchestrator.db import ProductBlockTable, ProductTable, db
from orchestrator.schemas import ProductSchema

router = APIRouter()


@router.get(
    "/",
    response_model=list[ProductSchema],
)
def fetch(tag: str | None = None, product_type: str | None = None) -> list[ProductSchema]:
    stmt = select(ProductTable).options(
        selectinload(ProductTable.workflows),
        selectinload(ProductTable.fixed_inputs),
        selectinload(ProductTable.product_blocks).selectinload(ProductBlockTable.resource_types),
    )
    if tag:
        stmt = stmt.filter(ProductTable.__dict__["tag"] == tag)
    if product_type:
        stmt = stmt.filter(ProductTable.__dict__["product_type"] == product_type)

    return list(db.session.scalars(stmt))


@router.get(
    "/{product_id}",
    response_model=ProductSchema,
)
def product_by_id(product_id: UUID) -> ProductTable:
    stmt = (
        select(ProductTable)
        .options(
            joinedload(ProductTable.fixed_inputs),
            joinedload(ProductTable.product_blocks),
            joinedload(ProductTable.workflows),
        )
        .filter(ProductTable.product_id == product_id)
    )
    product = db.session.scalars(stmt).unique().one_or_none()
    if not product:
        raise_status(HTTPStatus.NOT_FOUND, f"Product id {product_id} not found")
    return product



"""Create data head.

Revision ID: 59e1199aff7f
Revises:
Create Date: 2024-02-20T21:01:44.918799

"""

# revision identifiers, used by Alembic.
revision = "59e1199aff7f"
down_revision = None
branch_labels = ("data",)
depends_on = "048219045729"


def upgrade() -> None:
    pass


def downgrade() -> None:
    pass



"""Initial schema migration.

Revision ID: c112305b07d3
Revises:
Create Date: 2020-10-19 08:48:12.231037

"""

import sqlalchemy as sa
import sqlalchemy_utils
from alembic import op
from sqlalchemy.dialects import postgresql

from orchestrator import db

# revision identifiers, used by Alembic.
revision = "c112305b07d3"
down_revision = None
branch_labels = ("schema",)
depends_on = None


def upgrade() -> None:
    conn = op.get_bind()
    conn.execute(sa.text('CREATE EXTENSION IF NOT EXISTS "uuid-ossp" WITH SCHEMA public'))

    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "engine_settings",
        sa.Column("global_lock", sa.Boolean(), nullable=False),
        sa.Column("running_processes", sa.Integer(), nullable=False),
        sa.CheckConstraint("running_processes >= 0", name="check_running_processes_positive"),
        sa.PrimaryKeyConstraint("global_lock"),
    )
    op.create_table(
        "processes",
        sa.Column(
            "pid", sqlalchemy_utils.types.uuid.UUIDType(), server_default=sa.text("uuid_generate_v4()"), nullable=False
        ),
        sa.Column("workflow", sa.String(length=255), nullable=False),
        sa.Column("assignee", sa.String(length=50), server_default="SYSTEM", nullable=False),
        sa.Column("last_status", sa.String(length=50), nullable=False),
        sa.Column("last_step", sa.String(length=255), nullable=True),
        sa.Column(
            "started_at", db.UtcTimestamp(timezone=True), server_default=sa.text("current_timestamp"), nullable=False
        ),
        sa.Column(
            "last_modified_at",
            db.UtcTimestamp(timezone=True),
            server_default=sa.text("current_timestamp"),
            nullable=False,
        ),
        sa.Column("failed_reason", sa.Text(), nullable=True),
        sa.Column("traceback", sa.Text(), nullable=True),
        sa.Column("created_by", sa.String(length=255), nullable=True),
        sa.Column("is_task", sa.Boolean(), server_default=sa.text("false"), nullable=False),
        sa.PrimaryKeyConstraint("pid"),
    )
    op.create_index(op.f("ix_processes_is_task"), "processes", ["is_task"], unique=False)
    op.create_index(op.f("ix_processes_pid"), "processes", ["pid"], unique=False)
    op.create_table(
        "product_blocks",
        sa.Column(
            "product_block_id",
            sqlalchemy_utils.types.uuid.UUIDType(),
            server_default=sa.text("uuid_generate_v4()"),
            nullable=False,
        ),
        sa.Column("name", sa.String(), nullable=False),
        sa.Column("description", sa.Text(), nullable=False),
        sa.Column("tag", sa.String(length=20), nullable=True),
        sa.Column("status", sa.String(length=255), nullable=True),
        sa.Column(
            "created_at", db.UtcTimestamp(timezone=True), server_default=sa.text("current_timestamp"), nullable=False
        ),
        sa.Column("end_date", db.UtcTimestamp(timezone=True), nullable=True),
        sa.PrimaryKeyConstraint("product_block_id"),
        sa.UniqueConstraint("name"),
    )
    op.create_table(
        "products",
        sa.Column(
            "product_id",
            sqlalchemy_utils.types.uuid.UUIDType(),
            server_default=sa.text("uuid_generate_v4()"),
            nullable=False,
        ),
        sa.Column("name", sa.String(), nullable=False),
        sa.Column("description", sa.Text(), nullable=False),
        sa.Column("product_type", sa.String(length=255), nullable=False),
        sa.Column("tag", sa.String(length=20), nullable=False),
        sa.Column("status", sa.String(length=255), nullable=False),
        sa.Column(
            "created_at", db.UtcTimestamp(timezone=True), server_default=sa.text("current_timestamp"), nullable=False
        ),
        sa.Column("end_date", db.UtcTimestamp(timezone=True), nullable=True),
        sa.PrimaryKeyConstraint("product_id"),
        sa.UniqueConstraint("name"),
    )
    op.create_index(op.f("ix_products_tag"), "products", ["tag"], unique=False)
    op.create_table(
        "resource_types",
        sa.Column(
            "resource_type_id",
            sqlalchemy_utils.types.uuid.UUIDType(),
            server_default=sa.text("uuid_generate_v4()"),
            nullable=False,
        ),
        sa.Column("resource_type", sa.String(length=510), nullable=False),
        sa.Column("description", sa.Text(), nullable=True),
        sa.PrimaryKeyConstraint("resource_type_id"),
        sa.UniqueConstraint("resource_type"),
    )
    op.create_table(
        "workflows",
        sa.Column(
            "workflow_id",
            sqlalchemy_utils.types.uuid.UUIDType(),
            server_default=sa.text("uuid_generate_v4()"),
            nullable=False,
        ),
        sa.Column("name", sa.String(), nullable=False),
        sa.Column("target", sa.String(), nullable=False),
        sa.Column("description", sa.Text(), nullable=True),
        sa.Column(
            "created_at", db.UtcTimestamp(timezone=True), server_default=sa.text("current_timestamp"), nullable=False
        ),
        sa.PrimaryKeyConstraint("workflow_id"),
        sa.UniqueConstraint("name"),
    )
    op.create_table(
        "fixed_inputs",
        sa.Column(
            "fixed_input_id",
            sqlalchemy_utils.types.uuid.UUIDType(),
            server_default=sa.text("uuid_generate_v4()"),
            nullable=False,
        ),
        sa.Column("name", sa.String(), nullable=False),
        sa.Column("value", sa.String(), nullable=False),
        sa.Column(
            "created_at", sa.TIMESTAMP(timezone=True), server_default=sa.text("current_timestamp"), nullable=False
        ),
        sa.Column("product_id", sqlalchemy_utils.types.uuid.UUIDType(), nullable=False),
        sa.ForeignKeyConstraint(["product_id"], ["products.product_id"], ondelete="CASCADE"),
        sa.PrimaryKeyConstraint("fixed_input_id"),
        sa.UniqueConstraint("name", "product_id"),
    )
    op.create_table(
        "process_steps",
        sa.Column(
            "stepid",
            sqlalchemy_utils.types.uuid.UUIDType(),
            server_default=sa.text("uuid_generate_v4()"),
            nullable=False,
        ),
        sa.Column("pid", sqlalchemy_utils.types.uuid.UUIDType(), nullable=False),
        sa.Column("name", sa.String(), nullable=False),
        sa.Column("status", sa.String(length=50), nullable=False),
        sa.Column("state", postgresql.JSONB(astext_type=sa.Text()), nullable=False),  # type: ignore
        sa.Column("created_by", sa.String(length=255), nullable=True),
        sa.Column(
            "executed_at",
            db.UtcTimestamp(timezone=True),
            server_default=sa.text("statement_timestamp()"),
            nullable=False,
        ),
        sa.Column("commit_hash", sa.String(length=40), nullable=True),
        sa.ForeignKeyConstraint(["pid"], ["processes.pid"], ondelete="CASCADE"),
        sa.PrimaryKeyConstraint("stepid"),
    )
    op.create_index(op.f("ix_process_steps_pid"), "process_steps", ["pid"], unique=False)
    op.create_table(
        "product_block_resource_types",
        sa.Column("product_block_id", sqlalchemy_utils.types.uuid.UUIDType(), nullable=False),
        sa.Column("resource_type_id", sqlalchemy_utils.types.uuid.UUIDType(), nullable=False),
        sa.ForeignKeyConstraint(["product_block_id"], ["product_blocks.product_block_id"], ondelete="CASCADE"),
        sa.ForeignKeyConstraint(["resource_type_id"], ["resource_types.resource_type_id"], ondelete="CASCADE"),
        sa.PrimaryKeyConstraint("product_block_id", "resource_type_id"),
    )
    op.create_table(
        "product_product_blocks",
        sa.Column("product_id", sqlalchemy_utils.types.uuid.UUIDType(), nullable=False),
        sa.Column("product_block_id", sqlalchemy_utils.types.uuid.UUIDType(), nullable=False),
        sa.ForeignKeyConstraint(["product_block_id"], ["product_blocks.product_block_id"], ondelete="CASCADE"),
        sa.ForeignKeyConstraint(["product_id"], ["products.product_id"], ondelete="CASCADE"),
        sa.PrimaryKeyConstraint("product_id", "product_block_id"),
    )
    op.create_table(
        "products_workflows",
        sa.Column("product_id", sqlalchemy_utils.types.uuid.UUIDType(), nullable=False),
        sa.Column("workflow_id", sqlalchemy_utils.types.uuid.UUIDType(), nullable=False),
        sa.ForeignKeyConstraint(["product_id"], ["products.product_id"], ondelete="CASCADE"),
        sa.ForeignKeyConstraint(["workflow_id"], ["workflows.workflow_id"], ondelete="CASCADE"),
        sa.PrimaryKeyConstraint("product_id", "workflow_id"),
    )
    op.create_table(
        "subscriptions",
        sa.Column(
            "subscription_id",
            sqlalchemy_utils.types.uuid.UUIDType(),
            server_default=sa.text("uuid_generate_v4()"),
            nullable=False,
        ),
        sa.Column("description", sa.Text(), nullable=False),
        sa.Column("status", sa.String(length=255), nullable=False),
        sa.Column("product_id", sqlalchemy_utils.types.uuid.UUIDType(), nullable=False),
        sa.Column("customer_id", sqlalchemy_utils.types.uuid.UUIDType(), nullable=False),
        sa.Column("insync", sa.Boolean(), nullable=False),
        sa.Column("start_date", db.UtcTimestamp(timezone=True), nullable=True),
        sa.Column("end_date", db.UtcTimestamp(timezone=True), nullable=True),
        sa.Column("note", sa.Text(), nullable=True),
        sa.Column("tsv", sqlalchemy_utils.types.ts_vector.TSVectorType(), nullable=True),
        sa.ForeignKeyConstraint(["product_id"], ["products.product_id"]),
        sa.PrimaryKeyConstraint("subscription_id"),
    )
    op.create_index(op.f("ix_subscriptions_customer_id"), "subscriptions", ["customer_id"], unique=False)
    op.create_index(op.f("ix_subscriptions_product_id"), "subscriptions", ["product_id"], unique=False)
    op.create_index(op.f("ix_subscriptions_status"), "subscriptions", ["status"], unique=False)
    op.create_index("subscription_customer_ix", "subscriptions", ["subscription_id", "customer_id"], unique=False)
    op.create_index("subscription_product_ix", "subscriptions", ["subscription_id", "product_id"], unique=False)
    op.create_index("subscription_tsv_ix", "subscriptions", ["tsv"], unique=False, postgresql_using="gin")
    op.create_table(
        "processes_subscriptions",
        sa.Column(
            "id", sqlalchemy_utils.types.uuid.UUIDType(), server_default=sa.text("uuid_generate_v4()"), nullable=False
        ),
        sa.Column("pid", sqlalchemy_utils.types.uuid.UUIDType(), nullable=False),
        sa.Column("subscription_id", sqlalchemy_utils.types.uuid.UUIDType(), nullable=False),
        sa.Column(
            "created_at", db.UtcTimestamp(timezone=True), server_default=sa.text("current_timestamp"), nullable=False
        ),
        sa.Column("workflow_target", sa.String(length=255), server_default="CREATE", nullable=False),
        sa.ForeignKeyConstraint(["pid"], ["processes.pid"], ondelete="CASCADE"),
        sa.ForeignKeyConstraint(["subscription_id"], ["subscriptions.subscription_id"]),
        sa.PrimaryKeyConstraint("id"),
    )
    op.create_index(op.f("ix_processes_subscriptions_pid"), "processes_subscriptions", ["pid"], unique=False)
    op.create_index(
        op.f("ix_processes_subscriptions_subscription_id"), "processes_subscriptions", ["subscription_id"], unique=False
    )
    op.create_index("processes_subscriptions_ix", "processes_subscriptions", ["pid", "subscription_id"], unique=False)
    op.create_table(
        "subscription_customer_descriptions",
        sa.Column(
            "id", sqlalchemy_utils.types.uuid.UUIDType(), server_default=sa.text("uuid_generate_v4()"), nullable=False
        ),
        sa.Column("subscription_id", sqlalchemy_utils.types.uuid.UUIDType(), nullable=False),
        sa.Column("customer_id", sqlalchemy_utils.types.uuid.UUIDType(), nullable=False),
        sa.Column("description", sa.Text(), nullable=False),
        sa.Column(
            "created_at", db.UtcTimestamp(timezone=True), server_default=sa.text("current_timestamp"), nullable=False
        ),
        sa.ForeignKeyConstraint(["subscription_id"], ["subscriptions.subscription_id"], ondelete="CASCADE"),
        sa.PrimaryKeyConstraint("id"),
        sa.UniqueConstraint("customer_id", "subscription_id", name="uniq_customer_subscription_description"),
    )
    op.create_index(
        op.f("ix_subscription_customer_descriptions_customer_id"),
        "subscription_customer_descriptions",
        ["customer_id"],
        unique=False,
    )
    op.create_index(
        op.f("ix_subscription_customer_descriptions_subscription_id"),
        "subscription_customer_descriptions",
        ["subscription_id"],
        unique=False,
    )
    op.create_table(
        "subscription_instances",
        sa.Column(
            "subscription_instance_id",
            sqlalchemy_utils.types.uuid.UUIDType(),
            server_default=sa.text("uuid_generate_v4()"),
            nullable=False,
        ),
        sa.Column("subscription_id", sqlalchemy_utils.types.uuid.UUIDType(), nullable=False),
        sa.Column("product_block_id", sqlalchemy_utils.types.uuid.UUIDType(), nullable=False),
        sa.Column("label", sa.String(length=255), nullable=True),
        sa.ForeignKeyConstraint(["product_block_id"], ["product_blocks.product_block_id"]),
        sa.ForeignKeyConstraint(["subscription_id"], ["subscriptions.subscription_id"], ondelete="CASCADE"),
        sa.PrimaryKeyConstraint("subscription_instance_id"),
    )
    op.create_index(
        op.f("ix_subscription_instances_product_block_id"), "subscription_instances", ["product_block_id"], unique=False
    )
    op.create_index(
        op.f("ix_subscription_instances_subscription_id"), "subscription_instances", ["subscription_id"], unique=False
    )
    op.create_index(
        "subscription_instance_s_pb_ix",
        "subscription_instances",
        ["subscription_instance_id", "subscription_id", "product_block_id"],
        unique=False,
    )
    op.create_table(
        "subscription_instance_relations",
        sa.Column("parent_id", sqlalchemy_utils.types.uuid.UUIDType(), nullable=False),
        sa.Column("child_id", sqlalchemy_utils.types.uuid.UUIDType(), nullable=False),
        sa.Column("order_id", sa.Integer(), nullable=False),
        sa.Column("domain_model_attr", sa.Text(), nullable=True),
        sa.ForeignKeyConstraint(["child_id"], ["subscription_instances.subscription_instance_id"], ondelete="CASCADE"),
        sa.ForeignKeyConstraint(["parent_id"], ["subscription_instances.subscription_instance_id"], ondelete="CASCADE"),
        sa.PrimaryKeyConstraint("parent_id", "child_id", "order_id"),
    )
    op.create_index(
        "subscription_relation_p_c_o_ix",
        "subscription_instance_relations",
        ["parent_id", "child_id", "order_id"],
        unique=True,
    )
    op.create_table(
        "subscription_instance_values",
        sa.Column(
            "subscription_instance_value_id",
            sqlalchemy_utils.types.uuid.UUIDType(),
            server_default=sa.text("uuid_generate_v4()"),
            nullable=False,
        ),
        sa.Column("subscription_instance_id", sqlalchemy_utils.types.uuid.UUIDType(), nullable=False),
        sa.Column("resource_type_id", sqlalchemy_utils.types.uuid.UUIDType(), nullable=False),
        sa.Column("value", sa.Text(), nullable=False),
        sa.ForeignKeyConstraint(["resource_type_id"], ["resource_types.resource_type_id"]),
        sa.ForeignKeyConstraint(
            ["subscription_instance_id"], ["subscription_instances.subscription_instance_id"], ondelete="CASCADE"
        ),
        sa.PrimaryKeyConstraint("subscription_instance_value_id"),
    )
    op.create_index(
        op.f("ix_subscription_instance_values_resource_type_id"),
        "subscription_instance_values",
        ["resource_type_id"],
        unique=False,
    )
    op.create_index(
        op.f("ix_subscription_instance_values_subscription_instance_id"),
        "subscription_instance_values",
        ["subscription_instance_id"],
        unique=False,
    )
    op.create_index(
        "siv_si_rt_ix",
        "subscription_instance_values",
        ["subscription_instance_value_id", "subscription_instance_id", "resource_type_id"],
        unique=False,
    )
    # ### end Alembic commands ###

    conn.execute(
        sa.text(
            """
        CREATE TYPE public.tsq_state AS (
            search_query text,
            parentheses_stack integer,
            skip_for integer,
            current_token text,
            current_index integer,
            current_char text,
            previous_char text,
            tokens text[]
            )
        """
        )
    )

    conn.execute(
        sa.text(
            """
        CREATE FUNCTION public.array_nremove(anyarray, anyelement, integer) RETURNS anyarray
            LANGUAGE sql IMMUTABLE
            AS $_$
            WITH replaced_positions AS (
                SELECT UNNEST(
                    CASE
                    WHEN $2 IS NULL THEN
                        '{}'::int[]
                    WHEN $3 > 0 THEN
                        (array_positions($1, $2))[1:$3]
                    WHEN $3 < 0 THEN
                        (array_positions($1, $2))[
                            (cardinality(array_positions($1, $2)) + $3 + 1):
                        ]
                    ELSE
                        '{}'::int[]
                    END
                ) AS position
            )
            SELECT COALESCE((
                SELECT array_agg(value)
                FROM unnest($1) WITH ORDINALITY AS t(value, index)
                WHERE index NOT IN (SELECT position FROM replaced_positions)
            ), $1[1:0]);
        $_$;
        """
        )
    )
    conn.execute(
        sa.text(
            """
        CREATE FUNCTION public.fixed_inputs_trigger() RETURNS trigger
            LANGUAGE plpgsql
            AS $$
        BEGIN
            UPDATE subscriptions SET tsv = NULL WHERE product_id = NEW.product_id;
            RETURN NEW;
        END
        $$;
        """
        )
    )
    conn.execute(
        sa.text(
            """
        CREATE FUNCTION public.generate_subscription_tsv(sub_id uuid) RETURNS tsvector
            LANGUAGE plpgsql
            AS $$
        DECLARE
            gen_tsv TSVECTOR := NULL;
        BEGIN
            /*
             To generate a `tsvector` we first need to construct a `text` 'document' with
             meaningful data. That document is then fed into the `to_tsvector`
             function. That 'document' in our case boils down to all relevant data of a
             subscription. Because of our data model this requires a fair amount of
             joins. To make the data gathering step somewhat easier to follow we've
             split up the gathering of:

             - resource type data
             - subscription and product data
             - fixed input data
             - subscription customer description data

             into three separate queries using a Common Table Expression (CTE or WITH
             query). These three queries can be referenced to as tables, namely as
             tables:

             - rt_info
             - sub_prod_info
             - fi_info
             - cust_info

             The final query concatenates the corresponding rows of these three tables
             into a single document to be fed to `to_tsvector`.

             One thing of note is the usage of LEFT JOINs in the last query. This is
             for subscriptions that have been just created. Eg those that are 'initial'
             and as such might not yet have any resource types. A regular JOIN would
             not produce a result in that case leaving use with no document to feed
             into the `to_tsvector` function to populate the `tsv` column. Using a LEFT
             JOIN, together with `coalesce` to turn NULL values into empty strings
             (ensuring `concat_ws` does not return NULL), we will always get a
             meaningful value.
            */
            WITH rt_info AS (
                SELECT s.subscription_id,
                       string_agg(rt.resource_type || ': ' || siv.value, ', ' ORDER BY rt.resource_type) AS rt_info
                FROM subscription_instance_values siv
                         JOIN resource_types rt ON siv.resource_type_id = rt.resource_type_id
                         JOIN subscription_instances si ON siv.subscription_instance_id = si.subscription_instance_id
                         JOIN subscriptions s ON si.subscription_id = s.subscription_id
                GROUP BY s.subscription_id),
                 sub_prod_info AS (
                     SELECT s.subscription_id,
                            array_to_string(
                                    ARRAY ['subscription_id: ' || s.subscription_id,
                                        'status: ' || s.status,
                                        'insync: ' || s.insync,
                                        'subscription_description: ' || s.description,
                                        'note: ' || coalesce(s.note, ''),
                                        'customer_id: ' || s.customer_id,
                                        'product_id: ' || s.product_id],
                                    ', ') AS sub_info,
                            array_to_string(
                                    ARRAY ['product_name: ' || p.name,
                                        'product_description: ' || p.description,
                                        'tag: ' || p.tag,
                                        'product_type: ', p.product_type],
                                    ', ') AS prod_info
                     FROM subscriptions s
                              JOIN products p ON s.product_id = p.product_id),
                 fi_info AS (
                     SELECT s.subscription_id,
                            string_agg(fi.name || ': ' || fi.value, ', ' ORDER BY fi.name) AS fi_info
                     FROM subscriptions s
                              JOIN products p ON s.product_id = p.product_id
                              JOIN fixed_inputs fi ON p.product_id = fi.product_id
                     GROUP BY s.subscription_id),
                 cust_info AS (
                     SELECT s.subscription_id,
                            string_agg('customer_description: ' || scd.description, ', ') AS cust_info
                     FROM subscriptions s
                              JOIN subscription_customer_descriptions scd ON s.subscription_id = scd.subscription_id
                     GROUP BY s.subscription_id
                 )
            SELECT to_tsvector('english',
                           concat_ws(', ',
                                     coalesce(spi.sub_info, ''),
                                     coalesce(spi.prod_info, ''),
                                     coalesce(fi.fi_info, ''),
                                     coalesce(rti.rt_info, ''),
                                     coalesce(ci.cust_info, '')
                               ))
            INTO STRICT gen_tsv
            FROM subscriptions s
                     LEFT JOIN sub_prod_info spi ON s.subscription_id = spi.subscription_id
                     LEFT JOIN fi_info fi ON s.subscription_id = fi.subscription_id
                     LEFT JOIN rt_info rti ON s.subscription_id = rti.subscription_id
                     LEFT JOIN cust_info ci ON s.subscription_id = ci.subscription_id
            WHERE s.subscription_id = sub_id;
            RETURN gen_tsv;
        END
        $$;
        """
        )
    )

    conn.execute(
        sa.text(
            """
        CREATE FUNCTION public.products_trigger() RETURNS trigger
            LANGUAGE plpgsql
            AS $$
        BEGIN
            UPDATE subscriptions SET tsv = NULL WHERE product_id = NEW.product_id;
            RETURN NEW;
        END
        $$;
        """
        )
    )
    conn.execute(
        sa.text(
            """
        CREATE FUNCTION public.subscription_customer_descriptions_trigger() RETURNS trigger
            LANGUAGE plpgsql
            AS $$
        BEGIN
            UPDATE subscriptions SET tsv = NULL WHERE subscription_id = NEW.subscription_id;
            RETURN NEW;
        END
        $$;
        """
        )
    )
    conn.execute(
        sa.text(
            """
        CREATE FUNCTION public.subscription_instance_values_trigger() RETURNS trigger
            LANGUAGE plpgsql
            AS $$
        DECLARE
            sub_id subscriptions.subscription_id%TYPE;
        BEGIN
            SELECT si.subscription_id
            INTO STRICT sub_id
            FROM subscription_instances si
            WHERE si.subscription_instance_id = NEW.subscription_instance_id;
            UPDATE subscriptions SET tsv = NULL WHERE subscription_id = sub_id;
            RETURN NEW;
        END
        $$;
        """
        )
    )

    conn.execute(
        sa.text(
            """
        CREATE FUNCTION public.subscriptions_ins_trigger() RETURNS trigger
            LANGUAGE plpgsql
            AS $$
        BEGIN
            /*
             We have a separate insert trigger for subscriptions for the simple reason
             that while we are processing the insert there is not yet a row, with the
             just generated NEW.subscription_id, in the database to join with. Hence
             this trigger will only use the values from the insert.
             */
            SELECT to_tsvector('english',
                           concat_ws(', ',
                                     array_to_string(
                                             ARRAY ['subscription_id: ' || NEW.subscription_id,
                                                 'status: ' || NEW.status,
                                                 'insync: ' || NEW.insync,
                                                 'subscription_description: ' || NEW.description,
                                                 'note: ' || coalesce(NEW.note, ''),
                                                 'customer_id: ' || NEW.customer_id,
                                                 'product_id: ' || NEW.product_id],
                                             ', '),
                                     array_to_string(
                                             ARRAY ['product_name: ' || p.name,
                                                 'product_description: ' || p.description,
                                                 'tag: ' || p.tag,
                                                 'product_type: ', p.product_type],
                                             ', ')))
            INTO STRICT NEW.tsv
            FROM products p
            WHERE p.product_id = NEW.product_id;
            RETURN NEW;
        END
        $$;
        """
        )
    )

    conn.execute(
        sa.text(
            """
        CREATE FUNCTION public.subscriptions_upd_trigger() RETURNS trigger
            LANGUAGE plpgsql
            AS $$
        DECLARE
            upd_tsv TSVECTOR := NULL;
        BEGIN
            upd_tsv := generate_subscription_tsv(NEW.subscription_id);
            UPDATE subscriptions
            SET tsv = upd_tsv
            WHERE subscription_id = NEW.subscription_id
            AND tsv IS DISTINCT FROM upd_tsv;
            RETURN NULL;
        END
        $$;
        """
        )
    )

    conn.execute(
        sa.text(
            """
        CREATE FUNCTION public.tsq_append_current_token(state public.tsq_state) RETURNS public.tsq_state
            LANGUAGE plpgsql IMMUTABLE
            AS $$
        BEGIN
            IF state.current_token != '' THEN
                state.tokens := array_append(state.tokens, state.current_token);
                state.current_token := '';
            END IF;
            RETURN state;
        END;
        $$;
        """
        )
    )

    conn.execute(
        sa.text(
            """
        CREATE FUNCTION public.tsq_process_tokens(config regconfig, tokens text[]) RETURNS tsquery
            LANGUAGE plpgsql IMMUTABLE
            AS $$
        DECLARE
            result_query text;
            previous_value text;
            value text;
        BEGIN
            result_query := '';

            FOREACH value IN ARRAY tokens LOOP
                IF value = '"' THEN
                    CONTINUE;
                END IF;

                IF value = 'or' THEN
                    value := ' | ';
                END IF;

                IF left(value, 1) = '"' AND right(value, 1) = '"' THEN
                    value := phraseto_tsquery(config, value);
                ELSIF value NOT IN ('(', ' | ', ')', '-') THEN
                    value := quote_literal(value) || ':*';
                END IF;

                IF previous_value = '-' THEN
                    IF value = '(' THEN
                        value := '!' || value;
                    ELSIF value = ' | ' THEN
                        CONTINUE;
                    ELSE
                        value := '!(' || value || ')';
                    END IF;
                END IF;

                SELECT
                    CASE
                        WHEN result_query = '' THEN value
                        WHEN previous_value = ' | ' AND value = ' | ' THEN result_query
                        WHEN previous_value = ' | ' THEN result_query || ' | ' || value
                        WHEN previous_value IN ('!(', '(') OR value = ')' THEN result_query || value
                        WHEN value != ' | ' THEN result_query || ' & ' || value
                        ELSE result_query
                    END
                INTO result_query;
                previous_value := value;
            END LOOP;

            IF result_query = ' | ' THEN
                RETURN to_tsquery('');
            END IF;

            RETURN to_tsquery(config, result_query);
        END;
        $$;
        """
        )
    )

    conn.execute(
        sa.text(
            """
        CREATE FUNCTION public.tsq_tokenize(search_query text) RETURNS text[]
            LANGUAGE plpgsql IMMUTABLE
            AS $$
        DECLARE
            state tsq_state;
        BEGIN
            SELECT
                search_query::text AS search_query,
                0::int AS parentheses_stack,
                0 AS skip_for,
                ''::text AS current_token,
                0 AS current_index,
                ''::text AS current_char,
                ''::text AS previous_char,
                '{}'::text[] AS tokens
            INTO state;

            state.search_query := lower(trim(
                regexp_replace(search_query, '""+', '""', 'g')
            ));

            FOR state.current_index IN (
                SELECT generate_series(1, length(state.search_query))
            ) LOOP
                state.current_char := substring(
                    search_query FROM state.current_index FOR 1
                );

                IF state.skip_for > 0 THEN
                    state.skip_for := state.skip_for - 1;
                    CONTINUE;
                END IF;

                state := tsq_tokenize_character(state);
                state.previous_char := state.current_char;
            END LOOP;
            state := tsq_append_current_token(state);

            state.tokens := array_nremove(state.tokens, '(', -state.parentheses_stack);

            RETURN state.tokens;
        END;
        $$;
        """
        )
    )

    conn.execute(
        sa.text(
            """
        CREATE FUNCTION public.tsq_parse(config regconfig, search_query text) RETURNS tsquery
            LANGUAGE sql IMMUTABLE
            AS $$
            SELECT tsq_process_tokens(config, tsq_tokenize(search_query));
        $$;
        """
        )
    )

    conn.execute(
        sa.text(
            """
        CREATE FUNCTION public.tsq_parse(config text, search_query text) RETURNS tsquery
            LANGUAGE sql IMMUTABLE
            AS $$
            SELECT tsq_parse(config::regconfig, search_query);
        $$;
        """
        )
    )

    conn.execute(
        sa.text(
            """
        CREATE FUNCTION public.tsq_parse(search_query text) RETURNS tsquery
            LANGUAGE sql IMMUTABLE
            AS $$
            SELECT tsq_parse(get_current_ts_config(), search_query);
        $$;
        """
        )
    )

    conn.execute(
        sa.text(
            """
        CREATE FUNCTION public.tsq_process_tokens(tokens text[]) RETURNS tsquery
            LANGUAGE sql IMMUTABLE
            AS $$
            SELECT tsq_process_tokens(get_current_ts_config(), tokens);
        $$;
        """
        )
    )

    conn.execute(
        sa.text(
            """
        CREATE FUNCTION public.tsq_tokenize_character(state public.tsq_state) RETURNS public.tsq_state
            LANGUAGE plpgsql IMMUTABLE
            AS $$
        BEGIN
            IF state.current_char = '(' THEN
                state.tokens := array_append(state.tokens, '(');
                state.parentheses_stack := state.parentheses_stack + 1;
                state := tsq_append_current_token(state);
            ELSIF state.current_char = ')' THEN
                IF (state.parentheses_stack > 0 AND state.current_token != '') THEN
                    state := tsq_append_current_token(state);
                    state.tokens := array_append(state.tokens, ')');
                    state.parentheses_stack := state.parentheses_stack - 1;
                END IF;
            ELSIF state.current_char = '"' THEN
                state.skip_for := position('"' IN substring(
                    state.search_query FROM state.current_index + 1
                ));

                IF state.skip_for > 1 THEN
                    state.tokens = array_append(
                        state.tokens,
                        substring(
                            state.search_query
                            FROM state.current_index FOR state.skip_for + 1
                        )
                    );
                ELSIF state.skip_for = 0 THEN
                    state.current_token := state.current_token || state.current_char;
                END IF;
            ELSIF (
                state.current_char = '-' AND
                (state.current_index = 1 OR state.previous_char = ' ')
            ) THEN
                state.tokens := array_append(state.tokens, '-');
            ELSIF state.current_char = ' ' THEN
                state := tsq_append_current_token(state);
            ELSE
                state.current_token = state.current_token || state.current_char;
            END IF;
            RETURN state;
        END;
        $$;
        """
        )
    )

    conn.execute(
        sa.text(
            "CREATE TRIGGER fixed_inputs_trigger AFTER INSERT OR UPDATE ON public.fixed_inputs FOR EACH ROW EXECUTE FUNCTION public.fixed_inputs_trigger()"
        )
    )
    conn.execute(
        sa.text(
            "CREATE TRIGGER products_trigger AFTER INSERT OR UPDATE ON public.products FOR EACH ROW EXECUTE FUNCTION public.products_trigger()"
        )
    )
    conn.execute(
        sa.text(
            "CREATE TRIGGER subscription_customer_descriptions_trigger AFTER INSERT OR UPDATE ON public.subscription_customer_descriptions FOR EACH ROW EXECUTE FUNCTION public.subscription_customer_descriptions_trigger()"
        )
    )
    conn.execute(
        sa.text(
            "CREATE TRIGGER subscription_instance_values_trigger AFTER INSERT OR UPDATE ON public.subscription_instance_values FOR EACH ROW EXECUTE FUNCTION public.subscription_instance_values_trigger()"
        )
    )
    conn.execute(
        sa.text(
            "CREATE TRIGGER subscriptions_ins_trigger BEFORE INSERT ON public.subscriptions FOR EACH ROW EXECUTE FUNCTION public.subscriptions_ins_trigger()"
        )
    )
    conn.execute(
        sa.text(
            "CREATE TRIGGER subscriptions_upd_trigger AFTER UPDATE ON public.subscriptions FOR EACH ROW WHEN ((NOT (old.tsv IS DISTINCT FROM new.tsv))) EXECUTE FUNCTION public.subscriptions_upd_trigger()"
        )
    )

    conn.execute(sa.text("INSERT INTO public.engine_settings VALUES (false, 0)"))


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_index("siv_si_rt_ix", table_name="subscription_instance_values")
    op.drop_index(
        op.f("ix_subscription_instance_values_subscription_instance_id"), table_name="subscription_instance_values"
    )
    op.drop_index(op.f("ix_subscription_instance_values_resource_type_id"), table_name="subscription_instance_values")
    op.drop_table("subscription_instance_values")
    op.drop_index("subscription_relation_p_c_o_ix", table_name="subscription_instance_relations")
    op.drop_table("subscription_instance_relations")
    op.drop_index("subscription_instance_s_pb_ix", table_name="subscription_instances")
    op.drop_index(op.f("ix_subscription_instances_subscription_id"), table_name="subscription_instances")
    op.drop_index(op.f("ix_subscription_instances_product_block_id"), table_name="subscription_instances")
    op.drop_table("subscription_instances")
    op.drop_index(
        op.f("ix_subscription_customer_descriptions_subscription_id"), table_name="subscription_customer_descriptions"
    )
    op.drop_index(
        op.f("ix_subscription_customer_descriptions_customer_id"), table_name="subscription_customer_descriptions"
    )
    op.drop_table("subscription_customer_descriptions")
    op.drop_index("processes_subscriptions_ix", table_name="processes_subscriptions")
    op.drop_index(op.f("ix_processes_subscriptions_subscription_id"), table_name="processes_subscriptions")
    op.drop_index(op.f("ix_processes_subscriptions_pid"), table_name="processes_subscriptions")
    op.drop_table("processes_subscriptions")
    op.drop_index("subscription_tsv_ix", table_name="subscriptions")
    op.drop_index("subscription_product_ix", table_name="subscriptions")
    op.drop_index("subscription_customer_ix", table_name="subscriptions")
    op.drop_index(op.f("ix_subscriptions_status"), table_name="subscriptions")
    op.drop_index(op.f("ix_subscriptions_product_id"), table_name="subscriptions")
    op.drop_index(op.f("ix_subscriptions_customer_id"), table_name="subscriptions")
    op.drop_table("subscriptions")
    op.drop_table("products_workflows")
    op.drop_table("product_product_blocks")
    op.drop_table("product_block_resource_types")
    op.drop_index(op.f("ix_process_steps_pid"), table_name="process_steps")
    op.drop_table("process_steps")
    op.drop_table("fixed_inputs")
    op.drop_table("workflows")
    op.drop_table("resource_types")
    op.drop_index(op.f("ix_products_tag"), table_name="products")
    op.drop_table("products")
    op.drop_table("product_blocks")
    op.drop_index(op.f("ix_processes_pid"), table_name="processes")
    op.drop_index(op.f("ix_processes_is_task"), table_name="processes")
    op.drop_table("processes")
    op.drop_table("engine_settings")
    # ### end Alembic commands ###
